import re
from datetime import datetime
from datetime import timedelta
from datetime import timezone

import chardet
import click
from click.exceptions import BadParameter

from code42cli.logger import CliLogger
from code42cli.util import print_numbered_list


class AutoDecodedFile(click.File):
    """Attempts to autodetect file's encoding prior to normal click.File processing."""

    def convert(self, value, param, ctx):
        try:
            with open(value, "rb") as file:
                self.encoding = chardet.detect(file.read())["encoding"]
            if self.encoding is None:
                CliLogger().log_error(f"Failed to detect encoding of file: {value}")
        except Exception:
            pass  # we'll let click.File do it's own exception handling for the filepath

        return super().convert(value, param, ctx)


class FileOrString(AutoDecodedFile):
    """Declares a parameter to be a file (if the argument begins with `@`), otherwise accepts it as
    a string.
    """

    def __init__(self):
        super().__init__("r")

    def convert(self, value, param, ctx):
        if value.startswith("@") or value == "-":
            value = value.lstrip("@")
            file = super().convert(value, param, ctx)
            return file.read()
        else:
            return value


class MagicDate(click.ParamType):
    """Declares a parameter to be a 'magic' date string. Accepts an optional `round` argument
    which can be a function that takes a datetime and returns it rounded appropriately. This allows
    imprecise "day" input values (2020-01-01, 3d) to be rounded to the start or end of the day
    if needed. Accepts the following values as user input:

    timestamp formats:
        yyyy-MM-dd
        yyyy-MM-dd HH
        yyyy-MM-dd HH:MM
        yyyy-MM-dd HH:MM:SS

    short-string (day, hour, min) formats:
        30d
        24h
        15m

    and converts them to datetime objects.
    """

    TIMESTAMP_REGEX = re.compile(r"(\d{4}-\d{2}-\d{2})(?:$|T|\s+)([0-9:]+)?")
    MAGIC_TIME_REGEX = re.compile(r"(\d+)([dhmDHM])$")
    HELP_TEXT = (
        "Accepts a date/time in yyyy-MM-dd (UTC) or yyyy-MM-dd HH:MM:SS "
        "(UTC+24-hr time) format where the 'time' portion of the string "
        "can be partial (e.g. '2020-01-01 12' or '2020-01-01 01:15') or "
        "a 'short time' value representing days (30d), hours (24h) or "
        "minutes (15m) from the current time."
    )

    name = "magicdate"

    def __init__(self, rounding_func=None):
        self.round = rounding_func

    def get_metavar(self, param):
        return "[DATE|TIMESTAMP|SHORT_TIME]"

    def __repr__(self):
        return "MagicDate"

    def convert(self, value, param, ctx):
        timestamp_match = self.TIMESTAMP_REGEX.match(value)
        magic_match = self.MAGIC_TIME_REGEX.match(value)

        if timestamp_match:
            date, time = timestamp_match.groups()
            dt = self._get_dt_from_date_time_pair(date, time)
            if not time and callable(self.round):
                dt = self.round(dt)

        elif magic_match:
            num, period = magic_match.groups()
            dt = self._get_dt_from_magic_time_pair(num, period)
            if period == "d" and callable(self.round):
                dt = self.round(dt)

        else:
            self.fail(self.HELP_TEXT, param=param)

        return dt.replace(tzinfo=timezone.utc)

    @staticmethod
    def _get_dt_from_magic_time_pair(num, period):
        num = int(num)
        period = period.lower()
        if period == "d":
            delta = timedelta(days=num)
        elif period == "h":
            delta = timedelta(hours=num)
        elif period == "m":
            delta = timedelta(minutes=num)
        else:
            raise BadParameter(f"Couldn't parse magic time string: {num}{period}")
        return datetime.utcnow() - delta

    @staticmethod
    def _get_dt_from_date_time_pair(date, time):
        date_format = "%Y-%m-%d %H:%M:%S"
        if time:
            time = "{}:{}:{}".format(*time.split(":") + ["00", "00"])
        else:
            time = "00:00:00"
        date_string = f"{date} {time}"
        try:
            dt = datetime.strptime(date_string, date_format)
        except ValueError:
            raise BadParameter(f"Unable to parse date string: {date_string}.")
        else:
            return dt


class MapChoice(click.Choice):
    """Choice subclass that takes an extra map of additional 'valid' keys to map to correct
    choices list, allowing backward compatible choice changes. The extra keys don't show up
    in help text, but work when passed as a choice.
    """

    def __init__(self, choices, extras_map, **kwargs):
        self.extras_map = extras_map
        super().__init__(choices, **kwargs)

    def convert(self, value, param, ctx):
        if value in self.extras_map:
            value = self.extras_map[value]

        return super().convert(value, param, ctx)


class PromptChoice(click.ParamType):
    def __init__(self, choices):
        self.choices = choices

    def print_choices(self):
        print_numbered_list(self.choices)

    def convert(self, value, param, ctx):
        try:
            choice_index = int(value) - 1
            return self.choices[choice_index]
        except Exception:
            self.fail("Invalid choice", param=param)


class TOTP(click.ParamType):
    """Validates param to be a 6-digit integer, which is what all Code42 TOTP tokens will be."""

    def get_metavar(self, param):
        return "TEXT"

    def convert(self, value, param, ctx):
        try:
            int(value)
            assert len(value) == 6
            return value
        except Exception:
            raise BadParameter(
                f"TOTP tokens should be a 6-digit integer. '{value}' was provided."
            )



import click

import code42cli.options as opt
from code42cli.click_ext.groups import OrderedGroup
from code42cli.cmds.search import SendToCommand
from code42cli.cmds.search.cursor_store import AuditLogCursorStore
from code42cli.cmds.search.options import server_options
from code42cli.date_helper import convert_datetime_to_timestamp
from code42cli.options import checkpoint_option
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.output_formats import OutputFormatter
from code42cli.util import hash_event
from code42cli.util import parse_timestamp
from code42cli.util import warn_interrupt

EVENT_KEY = "events"
AUDIT_LOGS_KEYWORD = "audit-logs"


def _get_audit_logs_default_header():
    return {
        "timestamp": "Timestamp",
        "type$": "Type",
        "actorName": "ActorName",
        "actorIpAddress": "ActorIpAddress",
        "userName": "AffectedUser",
        "userId": "AffectedUserUID",
    }


begin_option = opt.begin_option(
    AUDIT_LOGS_KEYWORD,
    callback=lambda ctx, param, arg: convert_datetime_to_timestamp(arg),
)
end_option = opt.end_option(
    AUDIT_LOGS_KEYWORD,
    callback=lambda ctx, param, arg: convert_datetime_to_timestamp(arg),
)
filter_option_usernames = click.option(
    "--actor-username",
    required=False,
    help="Filter results by actor usernames.",
    multiple=True,
)
filter_option_user_ids = click.option(
    "--actor-user-id",
    required=False,
    help="Filter results by actor user IDs.",
    multiple=True,
)
filter_option_user_ip_addresses = click.option(
    "--actor-ip",
    required=False,
    help="Filter results by user IP addresses.",
    multiple=True,
)
filter_option_affected_user_ids = click.option(
    "--affected-user-id",
    required=False,
    help="Filter results by affected user IDs.",
    multiple=True,
)
filter_option_affected_usernames = click.option(
    "--affected-username",
    required=False,
    help="Filter results by affected usernames.",
    multiple=True,
)
filter_option_event_types = click.option(
    "--event-type",
    required=False,
    help="Filter results by event types (e.g. search_issued, user_registered, user_deactivated).",
    multiple=True,
)


def filter_options(f):
    f = filter_option_event_types(f)
    f = filter_option_usernames(f)
    f = filter_option_user_ids(f)
    f = filter_option_user_ip_addresses(f)
    f = filter_option_affected_user_ids(f)
    f = filter_option_affected_usernames(f)
    f = end_option(f)
    f = begin_option(f)
    return f


@click.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def audit_logs(state):
    """Get and send audit log event data."""
    # store cursor getter on the group state so shared --begin option can use it in validation
    state.cursor_getter = _get_audit_log_cursor_store


@audit_logs.command()
@click.argument("checkpoint-name")
@sdk_options()
def clear_checkpoint(state, checkpoint_name):
    """Remove the saved audit log checkpoint from `--use-checkpoint/-c` mode."""
    _get_audit_log_cursor_store(state.profile.name).delete(checkpoint_name)


@audit_logs.command()
@filter_options
@format_option
@checkpoint_option(AUDIT_LOGS_KEYWORD)
@sdk_options()
def search(
    state,
    begin,
    end,
    event_type,
    actor_username,
    actor_user_id,
    actor_ip,
    affected_user_id,
    affected_username,
    format,
    use_checkpoint,
):
    """Search audit log events."""
    formatter = OutputFormatter(format, _get_audit_logs_default_header())
    cursor = _get_audit_log_cursor_store(state.profile.name)
    if use_checkpoint:
        checkpoint_name = use_checkpoint
        checkpoint = cursor.get(checkpoint_name)
        if checkpoint is not None:
            begin = checkpoint

    events = _get_all_audit_log_events(
        state.sdk,
        begin_time=begin,
        end_time=end,
        event_types=event_type,
        usernames=actor_username,
        user_ids=actor_user_id,
        user_ip_addresses=actor_ip,
        affected_user_ids=affected_user_id,
        affected_usernames=affected_username,
    )

    if use_checkpoint:
        checkpoint_name = use_checkpoint
        events = list(
            _dedupe_checkpointed_events_and_store_updated_checkpoint(
                cursor, checkpoint_name, events
            )
        )

    if not events:
        click.echo("No results found.", err=True)
        return

    formatter.echo_formatted_list(events)


@audit_logs.command(cls=SendToCommand)
@filter_options
@checkpoint_option(AUDIT_LOGS_KEYWORD)
@server_options
@sdk_options()
def send_to(
    state,
    begin,
    end,
    event_type,
    actor_username,
    actor_user_id,
    actor_ip,
    affected_user_id,
    affected_username,
    use_checkpoint,
    **kwargs,
):
    """Send audit log events to the given server address in JSON format.

    HOSTNAME format: address:port where port is optional and defaults to 514.
    """
    cursor = _get_audit_log_cursor_store(state.profile.name)
    if use_checkpoint:
        checkpoint_name = use_checkpoint
        checkpoint = cursor.get(checkpoint_name)
        if checkpoint is not None:
            begin = checkpoint

    events = _get_all_audit_log_events(
        state.sdk,
        begin_time=begin,
        end_time=end,
        event_types=event_type,
        usernames=actor_username,
        user_ids=actor_user_id,
        user_ip_addresses=actor_ip,
        affected_user_ids=affected_user_id,
        affected_usernames=affected_username,
    )
    if use_checkpoint:
        checkpoint_name = use_checkpoint
        events = _dedupe_checkpointed_events_and_store_updated_checkpoint(
            cursor, checkpoint_name, events
        )
    with warn_interrupt():
        event = None
        for event in events:
            state.logger.info(event)
        if event is None:  # generator was empty
            click.echo("No results found.")


def _get_all_audit_log_events(sdk, **filter_args):
    response_gen = sdk.auditlogs.get_all(**filter_args)
    events = []
    try:
        responses = list(response_gen)
    except KeyError:
        # API endpoint (get_page) returns a response without events key when no records are found
        # e.g {"paginationRangeStartIndex": 10000, "paginationRangeEndIndex": 10000, "totalResultCount": 1593}
        # we can remove this check once PL-93211 is resolved and deployed.
        return events

    for response in responses:
        if EVENT_KEY in response.data:
            response_events = response.data.get(EVENT_KEY)
            events.extend(response_events)

    return sorted(events, key=lambda x: x.get("timestamp"))


def _dedupe_checkpointed_events_and_store_updated_checkpoint(
    cursor, checkpoint_name, events
):
    """De-duplicates events across checkpointed runs. Since using the timestamp of the last event
    processed as the `--begin` time of the next run causes the last event to show up again in the
    next results, we hash the last event(s) of each run and store those hashes in the cursor to
    filter out on the next run. It's also possible that two events have the exact same timestamp, so
    `checkpoint_events` needs to be a list of hashes so we can filter out everything that's actually
    been processed.
    """

    checkpoint_events = cursor.get_events(checkpoint_name)
    new_timestamp = None
    new_events = []
    for event in events:
        event_hash = hash_event(event)
        if event_hash not in checkpoint_events:
            if event["timestamp"] != new_timestamp:
                new_timestamp = event["timestamp"]
                new_events.clear()
            new_events.append(event_hash)
            yield event
            ts = parse_timestamp(new_timestamp)
            cursor.replace(checkpoint_name, ts)
            cursor.replace_events(checkpoint_name, new_events)


def _get_audit_log_cursor_store(profile_name):
    return AuditLogCursorStore(profile_name)



import os

import click

from code42cli.errors import LoggedCLIError
from code42cli.logger import get_main_cli_logger
from code42cli.worker import Worker

_logger = get_main_cli_logger()


class BulkCommandType:
    ADD = "add"
    REMOVE = "remove"

    def __iter__(self):
        return iter([self.ADD, self.REMOVE])


def write_template_file(path, columns):
    with open(path, "w", encoding="utf8") as new_file:
        new_file.write(",".join(columns))


def generate_template_cmd_factory(group_name, commands_dict, help_message=None):
    """Helper function that creates a `generate-template` click command that can be added to `bulk`
    sub-command groups.

    Args:
        `group_name`: a str representing the parent command group this is generating templates for.
        `commands_dict`: a dict of the commands with their column names. Keys are the cmd
            names that will become the `cmd` argument, and values are the list of column names for
            the csv.

            If a cmd takes a flat file, value should be a string indicating what item the flat file
            rows should contain.
    """
    help_message = (
        help_message
        or "Generate the CSV template needed for bulk adding/removing users."
    )

    @click.command(help=help_message)
    @click.argument("cmd", type=click.Choice(list(commands_dict)))
    @click.option(
        "--path",
        "-p",
        type=click.Path(dir_okay=False, resolve_path=True, writable=True),
        help="Write template file to specific file path/name.",
    )
    def generate_template(cmd, path):
        columns = commands_dict[cmd]
        if not path:
            filename = f"{group_name}_bulk_{cmd.replace('-', '_')}.csv"
            path = os.path.join(os.getcwd(), filename)
        write_template_file(path, columns)

    return generate_template


def run_bulk_process(
    row_handler, rows, progress_label=None, stats=None, raise_global_error=True
):
    """Runs a bulk process.

    Args:
        row_handler (callable): A callable that you define to process values from the row as
            either *args or **kwargs.
        rows (iterable): the rows to process.
        progress_label: a label that prints with the progress bar.
        stats (WorkerStats): Pass in WorkerStats if doing error handling outside of the worker.
        raise_global_error (bool): Set to False to *NOT* raise a CLI error if any rows fail.
            This is useful if doing error handling outside of the worker class.

    Returns:
        :class:`WorkerStats`: A class containing the successes and failures count.
    """
    processor = _create_bulk_processor(
        row_handler,
        rows,
        progress_label,
        stats=stats,
        raise_global_error=raise_global_error,
    )
    return processor.run()


def _create_bulk_processor(
    row_handler, rows, progress_label, stats=None, raise_global_error=True
):
    """A factory method to create the bulk processor, useful for testing purposes."""
    return BulkProcessor(
        row_handler,
        rows,
        progress_label=progress_label,
        stats=stats,
        raise_global_error=raise_global_error,
    )


class BulkProcessor:
    """A class for bulk processing a file.

    Args:
        row_handler (callable): A callable that you define to process values from the row as
            either *args or **kwargs. For example, if it's a csv file with header `prop_a,prop_b`
            and first row `1,test`, then `row_handler` should receive kwargs
            `prop_a: '1', prop_b: 'test'` when processing the first row. If it's a flat file, then
            `row_handler` only needs to take an extra arg.
    """

    def __init__(
        self,
        row_handler,
        rows,
        worker=None,
        progress_label=None,
        stats=None,
        raise_global_error=True,
    ):
        total = len(rows)
        self._rows = rows
        self._row_handler = row_handler
        self._progress_bar = click.progressbar(
            length=len(self._rows),
            item_show_func=self._show_stats,
            label=progress_label,
        )
        self._raise_global_error = raise_global_error
        self.__worker = worker or Worker(5, total, bar=self._progress_bar, stats=stats)
        self._stats = self.__worker.stats

    def run(self):
        """Processes the csv rows specified in the ctor, calling `self.row_handler` on each row."""
        self._stats.reset_results()
        for row in self._rows:
            self._process_row(row)
        self.__worker.wait()
        self._handle_if_errors()
        return self._stats._results

    def _process_row(self, row):
        self._process_csv_row(row)

    def _process_csv_row(self, row):
        # Removes problems from including extra columns. Error messages from out of order args
        # are more indicative this way too.
        row.pop(None, None)

        row_values = {key: val if val != "" else None for key, val in row.items()}
        self.__worker.do_async(
            lambda *args, **kwargs: self._handle_row(*args, **kwargs), **row_values
        )

    def _handle_row(self, *args, **kwargs):
        return self._row_handler(*args, **kwargs)

    def _show_stats(self, _):
        return str(self._stats)

    def _handle_if_errors(self):
        click.echo("")
        if self._stats.total_errors and self._raise_global_error:
            raise LoggedCLIError("Some problems occurred during bulk processing.")



from functools import lru_cache

from code42cli.errors import UserDoesNotExistError


@lru_cache(maxsize=None)
def get_user_id(sdk, username):
    """Returns the user's UID (referred to by `user_id` in detection lists).
    Raises `UserDoesNotExistError` if the user doesn't exist in the Code42 server.

    Args:
        sdk (py42.sdk.SDKClient): The py42 sdk.
        username (str or unicode): The username of the user to get an ID for.

    Returns:
         str: The user ID for the user with the given username.
    """
    users = sdk.users.get_by_username(username)["users"]
    if not users:
        raise UserDoesNotExistError(username)
    return users[0]["userUid"]



JSON_TO_CEF_MAP = {
    "actor": "suser",
    "cloudDriveId": "aid",
    "createTimestamp": "fileCreateTime",
    "deviceUid": "deviceExternalId",
    "deviceUserName": "suser",
    "domainName": "dvchost",
    "emailRecipients": "duser",
    "emailSender": "suser",
    "eventId": "externalId",
    "eventTimestamp": "end",
    "exposure": "reason",
    "fileCategory": "fileType",
    "fileName": "fname",
    "filePath": "filePath",
    "fileSize": "fsize",
    "insertionTimestamp": "rt",
    "md5Checksum": "fileHash",
    "modifyTimestamp": "fileModificationTime",
    "osHostName": "shost",
    "processName": "sproc",
    "processOwner": "spriv",
    "publicIpAddress": "src",
    "removableMediaBusType": "cs1",
    "removableMediaCapacity": "cn1",
    "removableMediaName": "cs3",
    "removableMediaSerialNumber": "cs4",
    "removableMediaVendor": "cs2",
    "sharedWith": "duser",
    "source": "sourceServiceName",
    "syncDestination": "destinationServiceName",
    "tabUrl": "request",
    "url": "filePath",
    "userUid": "suid",
    "windowTitle": "requestClientApplication",
}

CEF_CUSTOM_FIELD_NAME_MAP = {
    "cn1Label": "Code42AEDRemovableMediaCapacity",
    "cs1Label": "Code42AEDRemovableMediaBusType",
    "cs2Label": "Code42AEDRemovableMediaVendor",
    "cs3Label": "Code42AEDRemovableMediaName",
    "cs4Label": "Code42AEDRemovableMediaSerialNumber",
}

FILE_EVENT_TO_SIGNATURE_ID_MAP = {
    "CREATED": "C42200",
    "MODIFIED": "C42201",
    "DELETED": "C42202",
    "READ_BY_APP": "C42203",
    "EMAILED": "C42204",
}



import py42.sdk
import py42.settings
import py42.settings.debug as debug
import requests
from click import prompt
from click import secho
from py42.exceptions import Py42UnauthorizedError
from requests.exceptions import ConnectionError
from requests.exceptions import SSLError

from code42cli.click_ext.types import TOTP
from code42cli.errors import Code42CLIError
from code42cli.errors import LoggedCLIError
from code42cli.logger import get_main_cli_logger

py42.settings.items_per_page = 500

logger = get_main_cli_logger()


def create_sdk(profile, is_debug_mode, password=None, totp=None):
    if is_debug_mode:
        py42.settings.debug.level = debug.DEBUG
    if profile.ignore_ssl_errors == "True":
        secho(
            f"Warning: Profile '{profile.name}' has SSL verification disabled. "
            "Adding certificate verification is strongly advised.",
            fg="red",
            err=True,
        )
        requests.packages.urllib3.disable_warnings(
            requests.packages.urllib3.exceptions.InsecureRequestWarning
        )
        py42.settings.verify_ssl_certs = False
    password = password or profile.get_password()
    return _validate_connection(profile.authority_url, profile.username, password, totp)


def _validate_connection(authority_url, username, password, totp=None):
    try:
        return py42.sdk.from_local_account(authority_url, username, password, totp=totp)
    except SSLError as err:
        logger.log_error(err)
        raise LoggedCLIError(
            f"Problem connecting to {authority_url}, SSL certificate verification failed.\nUpdate profile with --disable-ssl-errors to bypass certificate checks (not recommended!)."
        )
    except ConnectionError as err:
        logger.log_error(err)
        raise LoggedCLIError(f"Problem connecting to {authority_url}.")
    except Py42UnauthorizedError as err:
        logger.log_error(err)
        if "LoginConfig: LOCAL_2FA" in str(err):
            if totp is None:
                totp = prompt(
                    "Multi-factor authentication required. Enter TOTP", type=TOTP()
                )
                return _validate_connection(authority_url, username, password, totp)
            else:
                raise Code42CLIError(
                    f"Invalid credentials or TOTP token for user {username}."
                )
        else:
            raise Code42CLIError(f"Invalid credentials for user {username}.")
    except Exception as err:
        logger.log_error(err)
        raise LoggedCLIError("Unknown problem validating connection.")



import itertools

from py42.sdk.queries.alerts.filters import DateObserved
from py42.sdk.queries.fileevents.filters import EventTimestamp
from py42.sdk.queries.fileevents.filters import ExposureType
from py42.sdk.queries.fileevents.filters import InsertionTimestamp
from py42.sdk.queries.query_filter import FilterGroup
from py42.sdk.queries.query_filter import QueryFilterTimestampField

from code42cli import errors
from code42cli.date_helper import verify_timestamp_order
from code42cli.logger import get_main_cli_logger
from code42cli.output_formats import OutputFormat

logger = get_main_cli_logger()


def convert_to_or_query(filter_groups):
    and_group = FilterGroup([], "AND")
    or_group = FilterGroup([], "OR")
    filters = itertools.chain.from_iterable([f.filter_list for f in filter_groups])
    for _filter in filters:
        if _is_exempt_filter(_filter):
            and_group.filter_list.append(_filter)
        else:
            or_group.filter_list.append(_filter)
    if and_group.filter_list:
        return [and_group, or_group]
    else:
        return [or_group]


def _is_exempt_filter(f):
    # exclude timestamp filters by default from "OR" queries
    # if other filters need to be exempt when building a query, append them to this list
    # can either be a `QueryFilter` subclass, or a composed `FilterGroup` if more precision on
    # is needed for which filters should be "AND"ed
    or_query_exempt_filters = [
        InsertionTimestamp,
        EventTimestamp,
        DateObserved,
        ExposureType.exists(),
    ]

    for exempt in or_query_exempt_filters:
        if isinstance(exempt, FilterGroup):
            if f in exempt:
                return True
            else:
                continue
        elif f.term == exempt._term:
            return True
    return False


def try_get_default_header(include_all, default_header, output_format):
    """Returns appropriate header based on include-all and output format. If returns None,
    the CLI format option will figure out the header based on the data keys."""
    output_header = None if include_all else default_header
    if output_format != OutputFormat.TABLE and include_all:
        err_text = "--include-all only allowed for Table output format."
        logger.log_error(err_text)
        raise errors.Code42CLIError(err_text)
    return output_header


def create_time_range_filter(filter_cls, begin_date=None, end_date=None):
    """Creates a filter using the given filter class (must be a subclass of
    :class:`py42.sdk.queries.query_filter.QueryFilterTimestampField`) and date args. Returns
    `None` if both begin_date and end_date args are `None`.

    Args:
        filter_cls: The class of filter to create. (must be a subclass of
          :class:`py42.sdk.queries.query_filter.QueryFilterTimestampField`)
        begin_date: The begin date for the range.
        end_date: The end date for the range.
    """
    if not issubclass(filter_cls, QueryFilterTimestampField):
        raise Exception("filter_cls must be a subclass of QueryFilterTimestampField")

    if begin_date and end_date:
        verify_timestamp_order(begin_date, end_date)
        return filter_cls.in_range(begin_date, end_date)

    elif begin_date and not end_date:
        return filter_cls.on_or_after(begin_date)

    elif end_date and not begin_date:
        return filter_cls.on_or_before(end_date)



import difflib
import platform
import re
from collections import OrderedDict

import click
from py42.exceptions import Py42ActiveLegalHoldError
from py42.exceptions import Py42CaseAlreadyHasEventError
from py42.exceptions import Py42CaseNameExistsError
from py42.exceptions import Py42CloudAliasCharacterLimitExceededError
from py42.exceptions import Py42CloudAliasLimitExceededError
from py42.exceptions import Py42DescriptionLimitExceededError
from py42.exceptions import Py42ForbiddenError
from py42.exceptions import Py42HTTPError
from py42.exceptions import Py42InvalidEmailError
from py42.exceptions import Py42InvalidPasswordError
from py42.exceptions import Py42InvalidRuleOperationError
from py42.exceptions import Py42InvalidUsernameError
from py42.exceptions import Py42LegalHoldNotFoundOrPermissionDeniedError
from py42.exceptions import Py42NotFoundError
from py42.exceptions import Py42OrgNotFoundError
from py42.exceptions import Py42TrustedActivityConflictError
from py42.exceptions import Py42TrustedActivityIdNotFound
from py42.exceptions import Py42TrustedActivityInvalidCharacterError
from py42.exceptions import Py42UpdateClosedCaseError
from py42.exceptions import Py42UserAlreadyAddedError
from py42.exceptions import Py42UsernameMustBeEmailError
from py42.exceptions import Py42UserNotOnListError
from py42.exceptions import Py42UserRiskProfileNotFound
from py42.exceptions import Py42WatchlistNotFound

from code42cli.errors import Code42CLIError
from code42cli.errors import LoggedCLIError
from code42cli.errors import UserDoesNotExistError
from code42cli.logger import get_main_cli_logger
from code42cli.logger.handlers import SyslogServerNetworkConnectionError

_DIFFLIB_CUT_OFF = 0.6


class ExceptionHandlingGroup(click.Group):
    """A `click.Group` subclass to add custom exception handling."""

    logger = get_main_cli_logger()
    _original_args = None

    def make_context(self, info_name, args, parent=None, **extra):

        # grab the original command line arguments for logging purposes
        self._original_args = " ".join(args)

        return super().make_context(info_name, args, parent=parent, **extra)

    def invoke(self, ctx):
        try:
            return super().invoke(ctx)

        except click.UsageError as err:
            self._suggest_cmd(err)

        except LoggedCLIError:
            raise

        except Code42CLIError as err:
            self.logger.log_error(str(err))
            raise

        except click.ClickException:
            raise

        except click.exceptions.Exit:
            raise

        except (
            UserDoesNotExistError,
            Py42UserAlreadyAddedError,
            Py42UserNotOnListError,
            Py42InvalidRuleOperationError,
            Py42LegalHoldNotFoundOrPermissionDeniedError,
            SyslogServerNetworkConnectionError,
            Py42CaseNameExistsError,
            Py42DescriptionLimitExceededError,
            Py42CaseAlreadyHasEventError,
            Py42UpdateClosedCaseError,
            Py42UsernameMustBeEmailError,
            Py42InvalidEmailError,
            Py42InvalidPasswordError,
            Py42InvalidUsernameError,
            Py42ActiveLegalHoldError,
            Py42OrgNotFoundError,
            Py42TrustedActivityConflictError,
            Py42TrustedActivityInvalidCharacterError,
            Py42TrustedActivityIdNotFound,
            Py42CloudAliasLimitExceededError,
            Py42CloudAliasCharacterLimitExceededError,
            Py42UserRiskProfileNotFound,
            Py42WatchlistNotFound,
            Py42NotFoundError,
        ) as err:
            msg = err.args[0]
            self.logger.log_error(msg)
            raise Code42CLIError(msg)

        except Py42ForbiddenError as err:
            self.logger.log_verbose_error(self._original_args, err.response.request)
            raise LoggedCLIError(
                "You do not have the necessary permissions to perform this task. "
                "Try using or creating a different profile."
            )

        except Py42HTTPError as err:
            self.logger.log_verbose_error(self._original_args, err.response.request)
            raise LoggedCLIError("Problem making request to server.")

        except UnicodeEncodeError:
            if platform.system() == "Windows":
                cmd = 'if using powershell: $ENV:PYTHONIOENCODING="utf-16"\nif using cmd.exe:    SET PYTHONIOENCODING="utf-16"'
            else:
                cmd = 'export PYTHONIOENCODING="utf-8"'
            raise Code42CLIError(
                f"Failed to handle unicode character using environment's detected encoding, try running the following:\n\n{cmd}\n\nand then re-run your `code42` command."
            )

        except OSError:
            raise

        except Exception:
            self.logger.log_verbose_error()
            raise LoggedCLIError("Unknown problem occurred.")

    @staticmethod
    def _suggest_cmd(usage_err):
        """Handles fuzzy suggestion of commands that are close to the bad command entered."""
        if usage_err.message is not None:
            match = re.match("No such command '(.*)'.", usage_err.message)
            if match:
                bad_arg = match.groups()[0]
                available_commands = list(usage_err.ctx.command.commands.keys())
                suggested_commands = difflib.get_close_matches(
                    bad_arg, available_commands, cutoff=_DIFFLIB_CUT_OFF
                )
                if not suggested_commands:
                    raise usage_err
                usage_err.message = (
                    f"No such command '{bad_arg}'. "
                    f"Did you mean {' or '.join(suggested_commands)}?"
                )
        raise usage_err


class OrderedGroup(click.Group):
    """A `click.Group` subclass that uses an `OrderedDict` to store commands so the help text lists
    them in the order they were defined/added to the group.
    """

    def __init__(self, name=None, commands=None, **attrs):
        super().__init__(name, commands, **attrs)
        # the registered subcommands by their exported names.
        self.commands = commands or OrderedDict()

    def list_commands(self, ctx):
        return self.commands


class ExtensionGroup(ExceptionHandlingGroup):
    """A helper click.Group for extension scripts. If only a single command is added to this group,
    that command will be the "default" and won't need to be explicitly passed as the first argument
    to the extension script.
    """

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def parse_args(self, ctx, args):
        if len(self.commands) == 1:
            cmd_name, cmd = next(iter(self.commands.items()))
            if not args or args[0] not in self.commands:
                self.commands = {"": cmd}
                args.insert(0, "")
        super().parse_args(ctx, args)



import csv

import click
from pandas import DataFrame
from py42.constants import WatchlistType
from py42.exceptions import Py42NotFoundError
from py42.exceptions import Py42WatchlistNotFound

from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.click_ext.options import incompatible_with
from code42cli.click_ext.types import AutoDecodedFile
from code42cli.errors import Code42CLIError
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.output_formats import DataFrameOutputFormatter


@click.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def watchlists(state):
    """Manage watchlist user memberships."""
    pass


@watchlists.command("list")
@format_option
@sdk_options()
def _list(state, format):
    """List all watchlists."""
    pages = state.sdk.watchlists.get_all()
    dfs = (DataFrame(page["watchlists"]) for page in pages)
    formatter = DataFrameOutputFormatter(format)
    formatter.echo_formatted_dataframes(dfs)


@watchlists.command()
@click.option(
    "--watchlist-id",
    help="ID of the watchlist.",
)
@click.option(
    "--watchlist-type",
    type=click.Choice(WatchlistType.choices()),
    help="Type of watchlist to list.",
    cls=incompatible_with("watchlist_id"),
)
@click.option(
    "--only-included-users",
    help="Restrict results to users explicitly added to watchlist via API or Console. "
    "Users added implicitly via group membership or other dynamic rule will not be listed.",
    is_flag=True,
)
@format_option
@sdk_options()
def list_members(state, watchlist_type, watchlist_id, only_included_users, format):
    """List all members on a given watchlist."""
    if not watchlist_id and not watchlist_type:
        raise click.ClickException("--watchlist-id OR --watchlist-type is required.")
    if watchlist_type:
        watchlist_id = state.sdk.watchlists._watchlists_service.watchlist_type_id_map[
            watchlist_type
        ]
    if only_included_users:
        pages = state.sdk.watchlists.get_all_included_users(watchlist_id)
        dfs = (DataFrame(page["includedUsers"]) for page in pages)
    else:
        pages = state.sdk.watchlists.get_all_watchlist_members(watchlist_id)
        dfs = (DataFrame(page["watchlistMembers"]) for page in pages)
    formatter = DataFrameOutputFormatter(format)
    formatter.echo_formatted_dataframes(dfs)


@watchlists.command()
@click.option(
    "--watchlist-id",
    help="ID of the watchlist.",
)
@click.option(
    "--watchlist-type",
    type=click.Choice(WatchlistType.choices()),
    help="Type of watchlist to add user to.",
    cls=incompatible_with("watchlist_id"),
)
@click.argument("user", metavar="[USER_ID|USERNAME]")
@sdk_options()
def add(state, watchlist_id, watchlist_type, user):
    """Add a user to a watchlist."""
    if not watchlist_id and not watchlist_type:
        raise click.ClickException("--watchlist-id OR --watchlist-type is required.")
    try:
        user = int(user)
    except ValueError:
        # assume username if `user` is not an int
        user = state.sdk.userriskprofile.get_by_username(user)["userId"]
    try:
        if watchlist_id:
            state.sdk.watchlists.add_included_users_by_watchlist_id(user, watchlist_id)
        elif watchlist_type:
            state.sdk.watchlists.add_included_users_by_watchlist_type(
                user, watchlist_type
            )
    except Py42WatchlistNotFound:
        raise
    except Py42NotFoundError:
        raise Code42CLIError(f"User ID {user} not found.")


@watchlists.command()
@click.option("--watchlist-id", help="ID of the watchlist.")
@click.option(
    "--watchlist-type",
    type=click.Choice(WatchlistType.choices()),
    help="Type of watchlist to remove user from.",
    cls=incompatible_with("watchlist_id"),
)
@click.argument("user", metavar="[USER_ID|USERNAME]")
@sdk_options()
def remove(state, watchlist_id, watchlist_type, user):
    """Remove a user from a watchlist."""
    if not watchlist_id and not watchlist_type:
        raise click.ClickException("--watchlist-id OR --watchlist-type is required.")
    try:
        user = int(user)
    except ValueError:
        # assume username if `user` is not an int
        user = state.sdk.userriskprofile.get_by_username(user)["userId"]
    try:
        if watchlist_id:
            state.sdk.watchlists.remove_included_users_by_watchlist_id(
                user, watchlist_id
            )
        elif watchlist_type:
            state.sdk.watchlists.remove_included_users_by_watchlist_type(
                user, watchlist_type
            )
    except Py42WatchlistNotFound:
        raise
    except Py42NotFoundError:
        raise Code42CLIError(f"User ID {user} not found.")


@watchlists.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def bulk(state):
    """Tools for executing bulk watchlist actions."""
    pass


watchlists_generate_template = generate_template_cmd_factory(
    group_name="watchlists",
    commands_dict={
        "add": ["watchlist_id", "watchlist_type", "user_id", "username"],
        "remove": ["watchlist_id", "watchlist_type", "user_id", "username"],
    },
)
bulk.add_command(watchlists_generate_template)


@bulk.command(
    name="add",
    help="Bulk add users to watchlists using a CSV file. Requires either a `watchlist_id` or "
    "`watchlist_type` column header to identify the watchlist, and either a `user_id` or "
    "`username` column header to identify the user to add.",
)
@click.argument(
    "csv_rows",
    metavar="CSV_FILE",
    type=AutoDecodedFile("r"),
    callback=lambda ctx, param, arg: csv.DictReader(arg),
)
@sdk_options()
def bulk_add(state, csv_rows):
    headers = csv_rows.fieldnames
    if "user_id" not in headers and "username" not in headers:
        raise Code42CLIError(
            "CSV requires either a `username` or `user_id` "
            "column to identify which users to add to watchlist."
        )
    if "watchlist_id" not in headers and "watchlist_type" not in headers:
        raise Code42CLIError(
            "CSV requires either a `watchlist_id` or `watchlist_type` "
            "column to identify which watchlist to add user to."
        )

    sdk = state.sdk

    def handle_row(watchlist_id=None, watchlist_type=None, user_id=None, username=None):
        if username and not user_id:
            user_id = sdk.userriskprofile.get_by_username(username)["userId"]
        if watchlist_id:
            sdk.watchlists.add_included_users_by_watchlist_id(user_id, watchlist_id)
        elif watchlist_type:
            choices = WatchlistType.choices()
            if watchlist_type not in choices:
                raise Code42CLIError(
                    f"Provided watchlist_type `{watchlist_type}` for username={username}, "
                    f"user_id={user_id} row is invalid. Must be one of: {','.join(choices)}"
                )
            sdk.watchlists.add_included_users_by_watchlist_type(user_id, watchlist_type)
        else:
            raise Code42CLIError(
                f"Row for username={username}, user_id={user_id} "
                "missing value for `watchlist_id` or `watchlist_type` columns."
            )

    run_bulk_process(
        handle_row,
        list(csv_rows),
        progress_label="Adding users to Watchlists:",
    )


@bulk.command(
    name="remove",
    help="Bulk remove users from watchlists using a CSV file. Requires either a `watchlist_id` or "
    "`watchlist_type` column header to identify the watchlist, and either a `user_id` or "
    "`username` header to identify the user to remove.",
)
@click.argument(
    "csv_rows",
    metavar="CSV_FILE",
    type=AutoDecodedFile("r"),
    callback=lambda ctx, param, arg: csv.DictReader(arg),
)
@sdk_options()
def bulk_remove(state, csv_rows):
    headers = csv_rows.fieldnames
    if "user_id" not in headers and "username" not in headers:
        raise Code42CLIError(
            "CSV requires either a `username` or `user_id` "
            "column to identify which users to remove from watchlist."
        )
    if "watchlist_id" not in headers and "watchlist_type" not in headers:
        raise Code42CLIError(
            "CSV requires either a `watchlist_id` or `watchlist_type` "
            "column to identify which watchlist to remove user from."
        )

    sdk = state.sdk

    def handle_row(watchlist_id=None, watchlist_type=None, user_id=None, username=None):
        if username and not user_id:
            user_id = sdk.userriskprofile.get_by_username(username)["userId"]
        if watchlist_id:
            sdk.watchlists.remove_included_users_by_watchlist_id(user_id, watchlist_id)
        elif watchlist_type:
            sdk.watchlists.remove_included_users_by_watchlist_type(
                user_id, watchlist_type
            )
        else:
            raise Code42CLIError(
                f"Row for username={username}, user_id={user_id} "
                "missing value for `watchlist_id` or `watchlist_type` columns."
            )

    run_bulk_process(
        handle_row,
        list(csv_rows),
        progress_label="Adding users to Watchlists:",
    )



import click

from code42cli.errors import Code42CLIError
from code42cli.logger import get_logger_for_server
from code42cli.logger.enums import ServerProtocol
from code42cli.output_formats import OutputFormat


def _try_get_logger_for_server(hostname, protocol, output_format, certs):
    try:
        return get_logger_for_server(hostname, protocol, output_format, certs)
    except Exception as err:
        raise Code42CLIError(
            f"Unable to connect to {hostname}. Failed with error: {err}."
        )


class SendToCommand(click.Command):
    def invoke(self, ctx):
        certs = ctx.params.get("certs")
        hostname = ctx.params.get("hostname")
        protocol = ctx.params.get("protocol")
        output_format = ctx.params.get("format", OutputFormat.RAW)
        ignore_cert_validation = ctx.params.get("ignore_cert_validation")
        _handle_incompatible_args(protocol, ignore_cert_validation, certs)

        if ignore_cert_validation:
            certs = "ignore"

        ctx.obj.logger = _try_get_logger_for_server(
            hostname, protocol, output_format, certs
        )
        return super().invoke(ctx)


def _handle_incompatible_args(protocol, ignore_cert_validation, certs):
    if protocol == ServerProtocol.TLS_TCP:
        return

    arg = None
    if ignore_cert_validation is not None:
        arg = "--ignore-cert-validation"
    elif certs is not None:
        arg = "--certs"
    if arg is not None:
        raise click.BadOptionUsage(
            arg, f"'{arg}' can only be used with '--protocol {ServerProtocol.TLS_TCP}'."
        )



import click
import py42.sdk.queries.alerts.filters as f
from py42.exceptions import Py42NotFoundError
from py42.sdk.queries.alerts.alert_query import AlertQuery
from py42.sdk.queries.alerts.filters import AlertState
from py42.sdk.queries.alerts.filters import RuleType
from py42.sdk.queries.alerts.filters import Severity
from py42.util import format_dict

import code42cli.cmds.search.options as searchopt
import code42cli.errors as errors
import code42cli.options as opt
from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.cmds.search import SendToCommand
from code42cli.cmds.search.cursor_store import AlertCursorStore
from code42cli.cmds.search.options import server_options
from code42cli.cmds.util import convert_to_or_query
from code42cli.cmds.util import create_time_range_filter
from code42cli.cmds.util import try_get_default_header
from code42cli.date_helper import convert_datetime_to_timestamp
from code42cli.date_helper import limit_date_range
from code42cli.enums import JsonOutputFormat
from code42cli.enums import OutputFormat
from code42cli.file_readers import read_csv_arg
from code42cli.options import format_option
from code42cli.output_formats import OutputFormatter
from code42cli.util import hash_event
from code42cli.util import parse_timestamp
from code42cli.util import warn_interrupt

ALERTS_KEYWORD = "alerts"
ALERT_PAGE_SIZE = 25

begin = opt.begin_option(
    ALERTS_KEYWORD,
    callback=lambda ctx, param, arg: convert_datetime_to_timestamp(
        limit_date_range(arg, max_days_back=90)
    ),
)
end = opt.end_option(ALERTS_KEYWORD)
checkpoint = opt.checkpoint_option(ALERTS_KEYWORD)
advanced_query = searchopt.advanced_query_option(ALERTS_KEYWORD)
severity_option = click.option(
    "--severity",
    multiple=True,
    type=click.Choice(Severity.choices()),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.is_in_filter(f.Severity),
    help="Filter alerts by severity. Defaults to returning all severities.",
)
filter_state_option = click.option(
    "--state",
    multiple=True,
    type=click.Choice(AlertState.choices()),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.is_in_filter(f.AlertState),
    help="Filter alerts by status. Defaults to returning all statuses.",
)
actor_option = click.option(
    "--actor",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.is_in_filter(f.Actor),
    help="Filter alerts by including the given actor(s) who triggered the alert. "
    "Arguments must match the actor's cloud alias exactly.",
)
actor_contains_option = click.option(
    "--actor-contains",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.contains_filter(f.Actor),
    help="Filter alerts by including actor(s) whose cloud alias contains the given string.",
)
exclude_actor_option = click.option(
    "--exclude-actor",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.not_in_filter(f.Actor),
    help="Filter alerts by excluding the given actor(s) who triggered the alert. "
    "Arguments must match actor's cloud alias exactly.",
)
exclude_actor_contains_option = click.option(
    "--exclude-actor-contains",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.not_contains_filter(f.Actor),
    help="Filter alerts by excluding actor(s) whose cloud alias contains the given string.",
)
rule_name_option = click.option(
    "--rule-name",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.is_in_filter(f.RuleName),
    help="Filter alerts by including the given rule name(s).",
)
exclude_rule_name_option = click.option(
    "--exclude-rule-name",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.not_in_filter(f.RuleName),
    help="Filter alerts by excluding the given rule name(s).",
)
rule_id_option = click.option(
    "--rule-id",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.is_in_filter(f.RuleId),
    help="Filter alerts by including the given rule id(s).",
)
exclude_rule_id_option = click.option(
    "--exclude-rule-id",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.not_in_filter(f.RuleId),
    help="Filter alerts by excluding the given rule id(s).",
)
rule_type_option = click.option(
    "--rule-type",
    multiple=True,
    type=click.Choice(RuleType.choices()),
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.is_in_filter(f.RuleType),
    help="Filter alerts by including the given rule type(s).",
)
exclude_rule_type_option = click.option(
    "--exclude-rule-type",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.not_in_filter(f.RuleType),
    help="Filter alerts by excluding the given rule type(s).",
)
description_option = click.option(
    "--description",
    multiple=True,
    cls=searchopt.AdvancedQueryAndSavedSearchIncompatible,
    callback=searchopt.contains_filter(f.Description),
    help="Filter alerts by description. Does fuzzy search by default.",
)
send_to_format_options = click.option(
    "-f",
    "--format",
    type=click.Choice(JsonOutputFormat(), case_sensitive=False),
    help="The output format of the result. Defaults to json format.",
    default=JsonOutputFormat.RAW,
)
alert_id_arg = click.argument("alert-id")
note_option = click.option("--note", help="A note to attach to the alert.")
update_state_option = click.option(
    "--state",
    help="The state to give to the alert.",
    type=click.Choice(AlertState.choices()),
)


def _get_default_output_header():
    return {
        "id": "Id",
        "name": "RuleName",
        "actor": "Username",
        "createdAt": "ObservedDate",
        "state": "State",
        "severity": "Severity",
        "description": "Description",
    }


def search_options(f):
    f = checkpoint(f)
    f = advanced_query(f)
    f = end(f)
    f = begin(f)
    return f


def filter_options(f):
    f = actor_option(f)
    f = actor_contains_option(f)
    f = exclude_actor_option(f)
    f = exclude_actor_contains_option(f)
    f = rule_name_option(f)
    f = exclude_rule_name_option(f)
    f = rule_id_option(f)
    f = exclude_rule_id_option(f)
    f = rule_type_option(f)
    f = exclude_rule_type_option(f)
    f = description_option(f)
    f = severity_option(f)
    f = filter_state_option(f)
    return f


@click.group(cls=OrderedGroup)
@opt.sdk_options(hidden=True)
def alerts(state):
    """Get and send alert data."""
    # store cursor getter on the group state so shared --begin option can use it in validation
    state.cursor_getter = _get_alert_cursor_store


@alerts.command()
@click.argument("checkpoint-name")
@opt.sdk_options()
def clear_checkpoint(state, checkpoint_name):
    """Remove the saved alert checkpoint from `--use-checkpoint/-c` mode."""
    _get_alert_cursor_store(state.profile.name).delete(checkpoint_name)


@alerts.command()
@filter_options
@search_options
@click.option(
    "--or-query", is_flag=True, cls=searchopt.AdvancedQueryAndSavedSearchIncompatible
)
@opt.sdk_options()
@click.option(
    "--include-all",
    default=False,
    is_flag=True,
    help="Display simple properties of the primary level of the nested response.",
)
@format_option
def search(
    cli_state,
    format,
    begin,
    end,
    advanced_query,
    use_checkpoint,
    or_query,
    include_all,
    **kwargs,
):
    """Search for alerts."""
    output_header = try_get_default_header(
        include_all, _get_default_output_header(), format
    )
    formatter = OutputFormatter(format, output_header)
    cursor = _get_alert_cursor_store(cli_state.profile.name) if use_checkpoint else None
    if use_checkpoint:
        checkpoint_name = use_checkpoint
        checkpoint = cursor.get(checkpoint_name)
        if checkpoint is not None:
            begin = checkpoint

    query = _construct_query(cli_state, begin, end, advanced_query, or_query)
    alerts_gen = cli_state.sdk.alerts.get_all_alert_details(query)

    if use_checkpoint:
        checkpoint_name = use_checkpoint
        # update checkpoint to alertId of last event retrieved
        alerts_gen = _dedupe_checkpointed_events_and_store_updated_checkpoint(
            cursor, checkpoint_name, alerts_gen
        )
    alerts_list = []
    for alert in alerts_gen:
        alerts_list.append(alert)
    if not alerts_list:
        click.echo("No results found.")
        return
    formatter.echo_formatted_list(alerts_list)


def _construct_query(state, begin, end, advanced_query, or_query):

    if advanced_query:
        state.search_filters = advanced_query
    else:
        if begin or end:
            state.search_filters.append(
                create_time_range_filter(f.DateObserved, begin, end)
            )
    if or_query:
        state.search_filters = convert_to_or_query(state.search_filters)
    query = AlertQuery(*state.search_filters)
    query.page_size = ALERT_PAGE_SIZE
    query.sort_direction = "asc"
    query.sort_key = "CreatedAt"
    return query


def _dedupe_checkpointed_events_and_store_updated_checkpoint(
    cursor, checkpoint_name, alerts_gen
):
    """De-duplicates events across checkpointed runs. Since using the timestamp of the last event
    processed as the `--begin` time of the next run causes the last event to show up again in the
    next results, we hash the last event(s) of each run and store those hashes in the cursor to
    filter out on the next run. It's also possible that two events have the exact same timestamp, so
    `checkpoint_events` needs to be a list of hashes so we can filter out everything that's actually
    been processed.
    """

    checkpoint_alerts = cursor.get_alerts(checkpoint_name)
    new_timestamp = None
    new_alerts = []
    for alert in alerts_gen:
        event_hash = hash_event(alert)
        if event_hash not in checkpoint_alerts:
            if alert[f.DateObserved._term] != new_timestamp:
                new_timestamp = alert[f.DateObserved._term]
                new_alerts.clear()
            new_alerts.append(event_hash)
            yield alert
            ts = parse_timestamp(new_timestamp)
            cursor.replace(checkpoint_name, ts)
            cursor.replace_alerts(checkpoint_name, new_alerts)


@alerts.command(cls=SendToCommand)
@filter_options
@search_options
@click.option(
    "--or-query", is_flag=True, cls=searchopt.AdvancedQueryAndSavedSearchIncompatible
)
@opt.sdk_options()
@server_options
@click.option(
    "--include-all",
    default=False,
    is_flag=True,
    help="Display simple properties of the primary level of the nested response.",
)
@send_to_format_options
def send_to(cli_state, begin, end, advanced_query, use_checkpoint, or_query, **kwargs):
    """Send alerts to the given server address.

    HOSTNAME format: address:port where port is optional and defaults to 514.
    """
    cursor = _get_cursor(cli_state, use_checkpoint)

    if use_checkpoint:
        checkpoint_name = use_checkpoint
        checkpoint = cursor.get(checkpoint_name)
        if checkpoint is not None:
            begin = checkpoint

    query = _construct_query(cli_state, begin, end, advanced_query, or_query)
    alerts_gen = cli_state.sdk.alerts.get_all_alert_details(query)

    if use_checkpoint:
        checkpoint_name = use_checkpoint
        alerts_gen = _dedupe_checkpointed_events_and_store_updated_checkpoint(
            cursor, checkpoint_name, alerts_gen
        )
    with warn_interrupt():
        alert = None
        for alert in alerts_gen:
            cli_state.logger.info(alert)
        if alert is None:  # generator was empty
            click.echo("No results found.")


def _get_cursor(state, use_checkpoint):
    return _get_alert_cursor_store(state.profile.name) if use_checkpoint else None


def _get_alert_cursor_store(profile_name):
    return AlertCursorStore(profile_name)


@alerts.command()
@opt.sdk_options()
@alert_id_arg
@click.option(
    "--include-observations", is_flag=True, help="View observations of the alert."
)
def show(state, alert_id, include_observations):
    """Display the details of a single alert."""
    formatter = OutputFormatter(OutputFormat.TABLE, _get_default_output_header())

    try:
        response = state.sdk.alerts.get_details(alert_id)
    except Py42NotFoundError:
        raise errors.Code42CLIError(f"No alert found with ID '{alert_id}'.")

    alert = response["alerts"][0]
    formatter.echo_formatted_list([alert])

    # Show note details
    note = alert.get("note")
    if note:
        click.echo("\nNote:\n")
        click.echo(format_dict(note))

    if include_observations:
        observations = alert.get("observations")
        if observations:
            click.echo("\nObservations:\n")
            click.echo(format_dict(observations))
        else:
            click.echo("\nNo observations found.")


@alerts.command()
@opt.sdk_options()
@alert_id_arg
@update_state_option
@note_option
def update(cli_state, alert_id, state, note):
    """Update alert information."""
    _update_alert(cli_state.sdk, alert_id, state, note)


@alerts.group(cls=OrderedGroup)
@opt.sdk_options(hidden=True)
def bulk(state):
    """Tools for executing bulk alert actions."""
    pass


UPDATE_ALERT_CSV_HEADERS = ["id", "state", "note"]
update_alerts_generate_template = generate_template_cmd_factory(
    group_name=ALERTS_KEYWORD,
    commands_dict={"update": UPDATE_ALERT_CSV_HEADERS},
    help_message="Generate the CSV template needed for bulk alert commands.",
)
bulk.add_command(update_alerts_generate_template)


@bulk.command(
    name="update",
    help=f"Bulk update alerts using a CSV file with format: {','.join(UPDATE_ALERT_CSV_HEADERS)}",
)
@opt.sdk_options()
@read_csv_arg(headers=UPDATE_ALERT_CSV_HEADERS)
def bulk_update(cli_state, csv_rows):
    """Bulk update alerts."""
    sdk = cli_state.sdk

    def handle_row(id, state, note):
        _update_alert(sdk, id, state, note)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Updating alerts:",
    )


def _update_alert(sdk, alert_id, alert_state, note):
    if alert_state:
        sdk.alerts.update_state(alert_state, [alert_id], note=note)
    elif note:
        sdk.alerts.update_note(alert_id, note)



import queue
from threading import Lock
from threading import Thread
from time import sleep

from py42.exceptions import Py42ForbiddenError
from py42.exceptions import Py42HTTPError

from code42cli.errors import Code42CLIError
from code42cli.logger import get_main_cli_logger


def create_worker_stats(total):
    return WorkerStats(total)


class WorkerStats:
    """Stats about the tasks that have run."""

    def __init__(self, total):
        self.total = total

    _total_processed = 0
    _total_errors = 0
    _results = []
    __total_processed_lock = Lock()
    __total_errors_lock = Lock()
    __results_lock = Lock()

    @property
    def total_processed(self):
        """The total number of tasks executed."""
        return self._total_processed

    @property
    def total_errors(self):
        """The amount of errors that occurred."""
        return self._total_errors

    @property
    def total_successes(self):
        val = self._total_processed - self._total_errors
        return val if val >= 0 else 0

    @property
    def results(self):
        return self._results

    def __str__(self):
        return f"{self.total_successes} succeeded, {self._total_errors} failed out of {self.total}"

    def increment_total_processed(self):
        """+1 to self.total_processed"""
        with self.__total_processed_lock:
            self._total_processed += 1

    def increment_total_errors(self):
        """+1 to self.total_errors"""
        with self.__total_errors_lock:
            self._total_errors += 1

    def add_result(self, result):
        """add a result to the list"""
        with self.__results_lock:
            self._results.append(result)

    def reset_results(self):
        with self.__results_lock:
            self._results = []


class Worker:
    def __init__(self, thread_count, expected_total, bar=None, stats=None):
        self._queue = queue.Queue()
        self._thread_count = thread_count
        self._bar = bar
        self._stats = stats or WorkerStats(expected_total)
        self._tasks = 0
        self.__started = False
        self.__start_lock = Lock()
        self._logger = get_main_cli_logger()

    def do_async(self, func, *args, **kwargs):
        """Execute the given func asynchronously given *args and **kwargs.

        Args:
            func (callable): The function to execute asynchronously.
            *args (iter): Positional args to pass to the function.
            **kwargs (dict): Key-value args to pass to the function.
        """
        if not self.__started:
            with self.__start_lock:
                if not self.__started:
                    self.__start()
                    self.__started = True
        self._queue.put({"func": func, "args": args, "kwargs": kwargs})
        self._tasks += 1

    @property
    def stats(self):
        """Stats about the tasks that have been executed, such as the total errors that occurred."""
        return self._stats

    def wait(self):
        """Wait for the tasks in the queue to complete. This should usually be called before
        program termination."""
        while self._stats.total_processed < self._tasks:
            sleep(0.5)

    def _process_queue(self):
        while True:
            try:
                task = self._queue.get()
                func = task["func"]
                args = task["args"]
                kwargs = task["kwargs"]
                self._stats.add_result(func(*args, **kwargs))
            except Code42CLIError as err:
                self._increment_total_errors()
                self._logger.log_error(err)
            except Py42ForbiddenError as err:
                self._increment_total_errors()
                self._logger.log_verbose_error(http_request=err.response.request)
                self._logger.log_error(
                    "You do not have the necessary permissions to perform this task. "
                    "Try using or creating a different profile."
                )
            except Py42HTTPError as err:
                self._increment_total_errors()
                self._logger.log_verbose_error(http_request=err.response.request)
            except Exception:
                self._increment_total_errors()
                self._logger.log_verbose_error()
            finally:
                self._stats.increment_total_processed()
                if self._bar:
                    self._bar.update(1)
                self._queue.task_done()

    def __start(self):
        for _ in range(0, self._thread_count):
            t = Thread(target=self._process_queue)
            t.daemon = True
            t.start()

    def _increment_total_errors(self):
        self._stats.increment_total_errors()



from getpass import getpass

import click
from click import echo
from click import secho

import code42cli.profile as cliprofile
from code42cli.click_ext.types import PromptChoice
from code42cli.click_ext.types import TOTP
from code42cli.errors import Code42CLIError
from code42cli.options import yes_option
from code42cli.profile import CREATE_PROFILE_HELP
from code42cli.sdk_client import create_sdk
from code42cli.util import does_user_agree


@click.group()
def profile():
    """Manage Code42 connection settings."""
    pass


debug_option = click.option(
    "-d",
    "--debug",
    is_flag=True,
    help="Turn on debug logging.",
)
totp_option = click.option(
    "--totp", help="TOTP token for multi-factor authentication.", type=TOTP()
)


def profile_name_arg(required=False):
    return click.argument("profile_name", required=required)


def name_option(required=False):
    return click.option(
        "-n",
        "--name",
        required=required,
        help="The name of the Code42 CLI profile to use when executing this command.",
    )


def server_option(required=False):
    return click.option(
        "-s",
        "--server",
        required=required,
        help="The URL you use to sign into Code42.",
    )


def username_option(required=False):
    return click.option(
        "-u",
        "--username",
        required=required,
        help="The username of the Code42 API user.",
    )


password_option = click.option(
    "--password",
    help="The password for the Code42 API user. If this option is omitted, interactive prompts "
    "will be used to obtain the password.",
)

disable_ssl_option = click.option(
    "--disable-ssl-errors",
    is_flag=True,
    help="For development purposes, do not validate the SSL certificates of Code42 servers. "
    "This is not recommended, except for specific scenarios like testing.",
    default=None,
)


@profile.command()
@profile_name_arg()
def show(profile_name):
    """Print the details of a profile."""
    c42profile = cliprofile.get_profile(profile_name)
    echo(f"\n{c42profile.name}:")
    echo(f"\t* username = {c42profile.username}")
    echo(f"\t* authority url = {c42profile.authority_url}")
    echo(f"\t* ignore-ssl-errors = {c42profile.ignore_ssl_errors}")
    if cliprofile.get_stored_password(c42profile.name) is not None:
        echo("\t* A password is set.")
    echo("")
    echo("")


@profile.command()
@name_option(required=True)
@server_option(required=True)
@username_option(required=True)
@password_option
@totp_option
@yes_option(hidden=True)
@disable_ssl_option
@debug_option
def create(name, server, username, password, disable_ssl_errors, debug, totp):
    """Create profile settings. The first profile created will be the default."""
    cliprofile.create_profile(name, server, username, disable_ssl_errors)
    password = password or _prompt_for_password(name)
    if password:
        _set_pw(name, password, debug, totp=totp)
    echo(f"Successfully created profile '{name}'.")


@profile.command()
@name_option()
@server_option()
@username_option()
@password_option
@totp_option
@disable_ssl_option
@debug_option
def update(name, server, username, password, disable_ssl_errors, debug, totp):
    """Update an existing profile."""
    c42profile = cliprofile.get_profile(name)

    if not server and not username and not password and disable_ssl_errors is None:
        raise click.UsageError(
            "Must provide at least one of `--username`, `--server`, `--password`, or "
            "`--disable-ssl-errors` when updating a profile."
        )

    cliprofile.update_profile(c42profile.name, server, username, disable_ssl_errors)
    if not password and not c42profile.has_stored_password:
        password = _prompt_for_password(c42profile.name)
    if password:
        _set_pw(name, password, debug, totp=totp)

    echo(f"Profile '{c42profile.name}' has been updated.")


@profile.command()
@profile_name_arg()
@debug_option
def reset_pw(profile_name, debug):
    """\b
    Change the stored password for a profile. Only affects what's stored in the local profile,
    does not make any changes to the Code42 user account."""
    password = getpass()
    profile_name_saved = _set_pw(profile_name, password, debug)
    echo(f"Password updated for profile '{profile_name_saved}'.")


@profile.command("list")
def _list():
    """Show all existing stored profiles."""
    profiles = cliprofile.get_all_profiles()
    if not profiles:
        raise Code42CLIError("No existing profile.", help=CREATE_PROFILE_HELP)
    for c42profile in profiles:
        echo(str(c42profile))


@profile.command()
@profile_name_arg()
def use(profile_name):
    """\b
    Set a profile as the default. If not providing a profile-name,
    prompts for a choice from a list of all profiles."""

    if not profile_name:
        _select_profile_from_prompt()
        return

    _set_default_profile(profile_name)


@profile.command()
@yes_option()
@profile_name_arg(required=True)
def delete(profile_name):
    """Deletes a profile and its stored password (if any)."""
    try:
        cliprofile.get_profile(profile_name)
    except Code42CLIError:
        raise Code42CLIError(f"Profile '{profile_name}' does not exist.")
    message = (
        "\nDeleting this profile will also delete any stored passwords and checkpoints. "
        "Are you sure? (y/n): "
    )
    if cliprofile.is_default_profile(profile_name):
        message = f"\n'{profile_name}' is currently the default profile!\n{message}"
    if does_user_agree(message):
        cliprofile.delete_profile(profile_name)
        echo(f"Profile '{profile_name}' has been deleted.")


@profile.command()
@yes_option()
def delete_all():
    """Deletes all profiles and saved passwords (if any)."""
    existing_profiles = cliprofile.get_all_profiles()
    if existing_profiles:
        profile_str_list = "\n\t".join(
            [c42profile.name for c42profile in existing_profiles]
        )
        message = (
            f"\nAre you sure you want to delete the following profiles?\n\t{profile_str_list}"
            "\n\nThis will also delete any stored passwords and checkpoints. (y/n): "
        )
        if does_user_agree(message):
            for profile_obj in existing_profiles:
                cliprofile.delete_profile(profile_obj.name)
                echo(f"Profile '{profile_obj.name}' has been deleted.")
    else:
        echo("\nNo profiles exist. Nothing to delete.")


def _prompt_for_password(profile_name):
    if does_user_agree("Would you like to set a password? (y/n): "):
        password = getpass()
        return password


def _set_pw(profile_name, password, debug, totp=None):
    c42profile = cliprofile.get_profile(profile_name)
    try:
        create_sdk(c42profile, is_debug_mode=debug, password=password, totp=totp)
    except Exception:
        secho("Password not stored!", bold=True)
        raise
    cliprofile.set_password(password, c42profile.name)
    return c42profile.name


def _select_profile_from_prompt():
    """Set the default profile from user input."""
    profiles = cliprofile.get_all_profiles()
    profile_names = [profile_choice.name for profile_choice in profiles]
    choices = PromptChoice(profile_names)
    choices.print_choices()
    prompt_message = "Input the number of the profile you wish to use"
    profile_name = click.prompt(prompt_message, type=choices)
    _set_default_profile(profile_name)


def _set_default_profile(profile_name):
    cliprofile.switch_default_profile(profile_name)
    _print_default_profile_was_set(profile_name)


def _print_default_profile_was_set(profile_name):
    echo(f"{profile_name} has been set as the default profile.")



import functools

import click
from pandas import DataFrame
from pandas import json_normalize
from py42.exceptions import Py42NotFoundError
from py42.exceptions import Py42UserRiskProfileNotFound

from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.click_ext.options import incompatible_with
from code42cli.errors import Code42CLIError
from code42cli.errors import UserDoesNotExistError
from code42cli.file_readers import read_csv_arg
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.output_formats import DataFrameOutputFormatter
from code42cli.output_formats import OutputFormat
from code42cli.output_formats import OutputFormatter
from code42cli.worker import create_worker_stats


username_arg = click.argument("username")

org_uid_option = click.option(
    "--org-uid",
    help="Limit users to only those in the organization you specify. Note that child orgs are included.",
)
active_option = click.option(
    "--active",
    is_flag=True,
    help="Limits results to only active users.",
    default=None,
)
inactive_option = click.option(
    "--inactive",
    is_flag=True,
    help="Limits results to only deactivated users.",
    cls=incompatible_with("active"),
)
user_id_option = click.option(
    "--user-id", help="The unique identifier of the user to be modified.", required=True
)
org_id_option = click.option(
    "--org-id",
    help="The unique identifier (UID) for the organization to which the user will be moved.",
    required=True,
)
include_legal_hold_option = click.option(
    "--include-legal-hold-membership",
    default=False,
    is_flag=True,
    help="Include legal hold membership in output.",
)


def role_name_option(help):
    return click.option("--role-name", help=help)


def username_option(help, required=False):
    return click.option("--username", help=help, required=required)


@click.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def users(state):
    """Manage users within your Code42 environment."""
    pass


@users.command(name="list")
@org_uid_option
@role_name_option("Limit results to only users having the specified role.")
@active_option
@inactive_option
@include_legal_hold_option
@click.option(
    "--include-roles", default=False, is_flag=True, help="Include user roles."
)
@format_option
@sdk_options()
def list_users(
    state,
    org_uid,
    role_name,
    active,
    inactive,
    include_legal_hold_membership,
    include_roles,
    format,
):
    """List users in your Code42 environment."""
    if inactive:
        active = False
    role_id = _get_role_id(state.sdk, role_name) if role_name else None
    columns = (
        ["userUid", "status", "username", "orgUid"]
        if format == OutputFormat.TABLE
        else None
    )
    if include_roles and columns:
        columns.append("roles")
    df = _get_users_dataframe(
        state.sdk, columns, org_uid, role_id, active, include_roles
    )
    if include_legal_hold_membership:
        df = _add_legal_hold_membership_to_user_dataframe(state.sdk, df)
    formatter = DataFrameOutputFormatter(format)
    formatter.echo_formatted_dataframes(df)


@users.command("show")
@username_arg
@include_legal_hold_option
@format_option
@sdk_options()
def show_user(state, username, include_legal_hold_membership, format):
    """Show user details."""
    columns = (
        ["userUid", "status", "username", "orgUid", "roles"]
        if format == OutputFormat.TABLE
        else None
    )
    response = state.sdk.users.get_by_username(username, incRoles=True)
    df = DataFrame.from_records(response["users"], columns=columns)
    if include_legal_hold_membership and not df.empty:
        df = _add_legal_hold_membership_to_user_dataframe(state.sdk, df)
    formatter = DataFrameOutputFormatter(format)
    formatter.echo_formatted_dataframes(df)


@users.command()
@username_option("Username of the target user.")
@role_name_option("Name of role to add.")
@sdk_options()
def add_role(state, username, role_name):
    """Add the specified role to the user with the specified username."""
    _add_user_role(state.sdk, username, role_name)


@users.command()
@role_name_option("Name of role to remove.")
@username_option("Username of the target user.")
@sdk_options()
def remove_role(state, username, role_name):
    """Remove the specified role to the user with the specified username."""
    _remove_user_role(state.sdk, role_name, username)


@users.command(name="update")
@user_id_option
@click.option("--username", help="The new username for the user.")
@click.option("--password", help="The new password for the user.")
@click.option("--email", help="The new email for the user.")
@click.option("--first-name", help="The new first name for the user.")
@click.option("--last-name", help="The new last name for the user.")
@click.option("--notes", help="Notes about this user.")
@click.option(
    "--archive-size-quota", help="The total size (in bytes) allowed for this user."
)
@sdk_options()
def update_user(
    state,
    user_id,
    username,
    email,
    password,
    first_name,
    last_name,
    notes,
    archive_size_quota,
):
    """Update a user with the specified unique identifier."""
    _update_user(
        state.sdk,
        user_id,
        username,
        email,
        password,
        first_name,
        last_name,
        notes,
        archive_size_quota,
    )


@users.command()
@username_arg
@sdk_options()
def deactivate(state, username):
    """Deactivate a user."""
    sdk = state.sdk
    _deactivate_user(sdk, username)


@users.command()
@username_arg
@sdk_options()
def reactivate(state, username):
    """Reactivate a user."""
    sdk = state.sdk
    _reactivate_user(sdk, username)


_bulk_user_update_headers = [
    "user_id",
    "username",
    "email",
    "password",
    "first_name",
    "last_name",
    "notes",
    "archive_size_quota",
]

_bulk_user_move_headers = ["username", "org_id"]

_bulk_user_roles_headers = ["username", "role_name"]

_bulk_user_alias_headers = ["username", "alias"]

_bulk_user_risk_profile_headers = ["username", "start_date", "end_date", "notes"]

_bulk_user_activation_headers = ["username"]


@users.command(name="move")
@username_option("The username of the user to move.", required=True)
@org_id_option
@sdk_options()
def change_organization(state, username, org_id):
    """Change the organization of the user with the given username
    to the org with the given org UID."""
    _change_organization(state.sdk, username, org_id)


@users.command()
@click.argument("username")
@click.argument("alias")
@sdk_options()
def add_alias(state, username, alias):
    """Add a cloud alias for a given user.

    A cloud alias is the username an employee uses to access cloud services such as Google Drive or Box. Adding a cloud alias allows Incydr to link a user's cloud activity with their Code42 username. Each user has a default cloud alias of their Code42 username. You can add one additional alias."""
    _add_cloud_alias(state.sdk, username, alias)


@users.command()
@click.argument("username")
@click.argument("alias")
@sdk_options()
def remove_alias(state, username, alias):
    """Remove a cloud alias for a given user."""
    _remove_cloud_alias(state.sdk, username, alias)


@users.command()
@click.argument("username")
@click.argument(
    "date", type=click.DateTime(formats=["%Y-%m-%d"]), required=False, metavar="DATE"
)
@click.option("--clear", is_flag=True, help="Clears the current `start_date` value.")
@sdk_options()
def update_start_date(state, username, date, clear):
    """Sets the `start_date` on a User's risk profile (useful for users on the New Hire Watchlist).
    Date format: %Y-%m-%d"""
    if not date and not clear:
        raise Code42CLIError("Must supply DATE argument if --clear is not used.")
    if clear:
        date = ""
    user_id = _get_user(state.sdk, username)["userId"]
    state.sdk.userriskprofile.update(user_id, start_date=date)


@users.command()
@click.argument("username")
@click.argument("date", type=click.DateTime(formats=["%Y-%m-%d"]), required=False)
@click.option("--clear", is_flag=True, help="Clears the current `end_date` value.")
@sdk_options()
def update_departure_date(state, username, date, clear):
    """Sets the `end_date` on a User's risk profile (useful for users on the Departing Watchlist). Date format: %Y-%m-%d"""
    if not date and not clear:
        raise Code42CLIError("Must supply DATE argument if --clear is not used.")
    if clear:
        date = ""
    user_id = _get_user(state.sdk, username)["userId"]
    state.sdk.userriskprofile.update(user_id, end_date=date)


@users.command()
@click.argument("username")
@click.argument("note", required=False)
@click.option("--clear", is_flag=True, help="Clears the current `notes` value.")
@click.option(
    "--append",
    is_flag=True,
    help="Appends provided note to existing note text as a new line.",
)
@sdk_options()
def update_risk_profile_notes(state, username, note, clear, append):
    """Sets the `notes` value of a User's risk profile.

    WARNING: Overwrites any existing note value."""
    if not note and not clear:
        raise Code42CLIError("Must supply NOTE argument if --clear is not used.")
    user = _get_user(state.sdk, username)
    user_id = user["userId"]
    if append and user["notes"]:
        note = user["notes"] + f"\n\n{note}"
    if clear:
        note = ""
    state.sdk.userriskprofile.update(user_id, notes=note)


@users.command()
@click.argument("username")
@sdk_options()
def list_aliases(state, username):
    """List the cloud aliases for a given user.

    Each user has a default cloud alias of their Code42 username with up to one additional alias."""
    user = _get_user(state.sdk, username)
    aliases = user["cloudAliases"]
    if aliases:
        click.echo(aliases)
    else:
        click.echo(f"No cloud aliases for user '{username}' found.")


@users.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def orgs(state):
    """Tools for viewing user orgs."""
    pass


def _get_orgs_header():
    return {
        "orgId": "ID",
        "orgUid": "UID",
        "orgName": "Name",
        "status": "Status",
        "parentOrgId": "Parent ID",
        "parentOrgUid": "Parent UID",
        "type": "Type",
        "classification": "Classification",
        "creationDate": "Creation Date",
        "settings": "Settings",
    }


@orgs.command(name="list")
@format_option
@sdk_options()
def list_orgs(
    state,
    format,
):
    """List all orgs."""
    pages = state.sdk.orgs.get_all()
    formatter = OutputFormatter(format, _get_orgs_header())
    orgs = [org for page in pages for org in page["orgs"]]
    if orgs:
        formatter.echo_formatted_list(orgs)
    else:
        click.echo("No orgs found.")


@orgs.command(name="show")
@click.argument("org-uid")
@format_option
@sdk_options()
def show_org(
    state,
    org_uid,
    format,
):
    """Show org details."""
    formatter = OutputFormatter(format)
    try:
        response = state.sdk.orgs.get_by_uid(org_uid)
        formatter.echo_formatted_list([response.data])
    except Py42NotFoundError:
        raise Code42CLIError(f"Invalid org UID {org_uid}.")


@users.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def bulk(state):
    """Tools for managing users in bulk."""
    pass


users_generate_template = generate_template_cmd_factory(
    group_name="users",
    commands_dict={
        "update": _bulk_user_update_headers,
        "move": _bulk_user_move_headers,
        "add-alias": _bulk_user_alias_headers,
        "remove-alias": _bulk_user_alias_headers,
        "update-risk-profile": _bulk_user_risk_profile_headers,
    },
    help_message="Generate the CSV template needed for bulk user commands.",
)
bulk.add_command(users_generate_template)


@bulk.command(
    name="update",
    help="Update a list of users from the provided CSV in format: "
    f"{','.join(_bulk_user_update_headers)}",
)
@read_csv_arg(headers=_bulk_user_update_headers)
@format_option
@sdk_options()
def bulk_update(state, csv_rows, format):
    """Update a list of users from the provided CSV."""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk

    csv_rows[0]["updated"] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _update_user(
                sdk, **{key: row[key] for key in row.keys() if key != "updated"}
            )
            row["updated"] = "True"
        except Exception as err:
            row["updated"] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Updating users:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(
    name="move",
    help="Change the organization of the list of users from the provided CSV in format: "
    f"{','.join(_bulk_user_move_headers)}",
)
@read_csv_arg(headers=_bulk_user_move_headers)
@format_option
@sdk_options()
def bulk_move(state, csv_rows, format):
    """Change the organization of the list of users from the provided CSV."""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk

    csv_rows[0]["moved"] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _change_organization(
                sdk, **{key: row[key] for key in row.keys() if key != "moved"}
            )
            row["moved"] = "True"
        except Exception as err:
            row["moved"] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Moving users:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(
    name="deactivate",
    help=f"Deactivate a list of users from the provided CSV in format: {','.join(_bulk_user_activation_headers)}",
)
@read_csv_arg(headers=_bulk_user_activation_headers)
@format_option
@sdk_options()
def bulk_deactivate(state, csv_rows, format):
    """Deactivate a list of users."""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk

    csv_rows[0]["deactivated"] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _deactivate_user(
                sdk, **{key: row[key] for key in row.keys() if key != "deactivated"}
            )
            row["deactivated"] = "True"
        except Exception as err:
            row["deactivated"] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Deactivating users:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(
    name="reactivate",
    help=f"Reactivate a list of users from the provided CSV in format: {','.join(_bulk_user_activation_headers)}",
)
@read_csv_arg(headers=_bulk_user_activation_headers)
@format_option
@sdk_options()
def bulk_reactivate(state, csv_rows, format):
    """Reactivate a list of users."""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk

    csv_rows[0]["reactivated"] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _reactivate_user(
                sdk, **{key: row[key] for key in row.keys() if key != "reactivated"}
            )
            row["reactivated"] = "True"
        except Exception as err:
            row["reactivated"] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Reactivating users:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(
    name="add-roles",
    help=f"Add roles to a list of users from the provided CSV in format: {','.join(_bulk_user_roles_headers)}",
)
@read_csv_arg(headers=_bulk_user_roles_headers)
@format_option
@sdk_options()
def bulk_add_roles(state, csv_rows, format):
    """Bulk add roles to a list of users."""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk
    status_header = "role added"

    csv_rows[0][status_header] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _add_user_role(
                sdk, **{key: row[key] for key in row.keys() if key != status_header}
            )
            row[status_header] = "True"
        except Exception as err:
            row[status_header] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Adding roles to users:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(
    name="remove-roles",
    help=f"Remove roles from a list of users from the provided CSV in format: {','.join(_bulk_user_roles_headers)}",
)
@read_csv_arg(headers=_bulk_user_roles_headers)
@format_option
@sdk_options()
def bulk_remove_roles(state, csv_rows, format):
    """Bulk remove roles from a list of users."""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk
    success_header = "role removed"

    csv_rows[0][success_header] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _remove_user_role(
                sdk, **{key: row[key] for key in row.keys() if key != success_header}
            )
            row[success_header] = "True"
        except Exception as err:
            row[success_header] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Removing roles from users:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(
    name="add-alias",
    help=f"Add aliases to a list of users from the provided CSV in format: {','.join(_bulk_user_alias_headers)}.\n\nA cloud alias is the username an employee uses to access cloud services such as Google Drive or Box. Adding a cloud alias allows Incydr to link a user's cloud activity with their Code42 username. Each user has a default cloud alias of their Code42 username. You can add one additional alias.",
)
@read_csv_arg(headers=_bulk_user_alias_headers)
@format_option
@sdk_options()
def bulk_add_alias(state, csv_rows, format):
    """Bulk add aliases to users"""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk
    success_header = "alias added"

    csv_rows[0][success_header] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _add_cloud_alias(
                sdk, **{key: row[key] for key in row.keys() if key != success_header}
            )
            row[success_header] = "True"
        except Exception as err:
            row[success_header] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Adding aliases to users:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(
    name="remove-alias",
    help=f"Remove aliases from a list of users from the provided CSV in format: {','.join(_bulk_user_alias_headers)}",
)
@read_csv_arg(headers=_bulk_user_alias_headers)
@format_option
@sdk_options()
def bulk_remove_alias(state, csv_rows, format):
    """Bulk remove aliases from users"""

    # Initialize the SDK before starting any bulk processes
    # to prevent multiple instances and having to enter 2fa multiple times.
    sdk = state.sdk
    success_header = "alias removed"

    csv_rows[0][success_header] = "False"
    formatter = OutputFormatter(format, {key: key for key in csv_rows[0].keys()})
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            _remove_cloud_alias(
                sdk, **{key: row[key] for key in row.keys() if key != success_header}
            )
            row[success_header] = "True"
        except Exception as err:
            row[success_header] = f"False: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Removing aliases from users:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


@bulk.command(
    name="update-risk-profile",
    help=f"Update user risk profile data from the provided CSV in format: {','.join(_bulk_user_risk_profile_headers)}"
    "\n\nTo clear a value, set column item to the string: 'null'.",
)
@format_option
@read_csv_arg(headers=_bulk_user_risk_profile_headers)
@click.option(
    "--append-notes",
    is_flag=True,
    help="Append provided note value to already existing note on a new line. Defaults to overwrite.",
)
@sdk_options()
def bulk_update_risk_profile(state, csv_rows, format, append_notes):
    """Bulk update User Risk Profile data."""
    sdk = state.sdk

    success_header = "updated_user"
    formatter = OutputFormatter(
        format, {key: key for key in [*csv_rows[0].keys(), success_header]}
    )
    stats = create_worker_stats(len(csv_rows))

    def handle_row(**row):
        try:
            updated_user = _update_userriskprofile(
                sdk, append_notes=append_notes, **row
            )
            row[success_header] = updated_user
        except Exception as err:
            row[success_header] = f"Error: {err}"
            stats.increment_total_errors()
        return row

    result_rows = run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Updating user risk profile data:",
        stats=stats,
        raise_global_error=False,
    )
    formatter.echo_formatted_list(result_rows)


def _add_user_role(sdk, username, role_name):
    user_id = _get_legacy_user_id(sdk, username)
    _get_role_id(sdk, role_name)  # function provides role name validation
    sdk.users.add_role(user_id, role_name)


def _remove_user_role(sdk, role_name, username):
    user_id = _get_legacy_user_id(sdk, username)
    _get_role_id(sdk, role_name)  # function provides role name validation
    sdk.users.remove_role(user_id, role_name)


def _get_legacy_user_id(sdk, username):
    if not username:
        # py42 returns all users when passing `None` to `get_by_username()`.
        raise click.BadParameter("Username is required.")
    user = sdk.users.get_by_username(username)["users"]
    if len(user) == 0:
        raise UserDoesNotExistError(username)
    user_id = user[0]["userId"]
    return user_id


@functools.lru_cache()
def _get_role_id(sdk, role_name):
    try:
        roles_dataframe = DataFrame.from_records(
            sdk.users.get_available_roles().data, index="roleName"
        )
        role_result = roles_dataframe.at[role_name, "roleId"]
        return str(role_result)  # extract the role ID from the series
    except KeyError:
        raise Code42CLIError(f"Role with name '{role_name}' not found.")


def _get_users_dataframe(sdk, columns, org_uid, role_id, active, include_roles):
    users_generator = sdk.users.get_all(
        active=active, org_uid=org_uid, role_id=role_id, incRoles=include_roles
    )
    users_list = []
    for page in users_generator:
        users_list.extend(page["users"])

    return DataFrame.from_records(users_list, columns=columns)


def _add_legal_hold_membership_to_user_dataframe(sdk, df):
    columns = ["legalHold.legalHoldUid", "legalHold.name", "user.userUid"]

    custodians = list(_get_all_active_hold_memberships(sdk))

    if len(custodians) == 0:
        return df

    legal_hold_member_dataframe = (
        json_normalize(custodians)[columns]
        .groupby(["user.userUid"])
        .agg(",".join)
        .rename(
            {
                "legalHold.legalHoldUid": "legalHoldUid",
                "legalHold.name": "legalHoldName",
            },
            axis=1,
        )
    )
    df = df.merge(
        legal_hold_member_dataframe,
        how="left",
        left_on="userUid",
        right_on="user.userUid",
    )

    return df


def _get_all_active_hold_memberships(sdk):
    for page in sdk.legalhold.get_all_matters(active=True):
        for matter in page["legalHolds"]:
            for _page in sdk.legalhold.get_all_matter_custodians(
                legal_hold_uid=matter["legalHoldUid"], active=True
            ):
                yield from _page["legalHoldMemberships"]


def _update_user(
    sdk,
    user_id,
    username,
    email,
    password,
    first_name,
    last_name,
    notes,
    archive_size_quota,
):
    return sdk.users.update_user(
        user_id,
        username=username,
        email=email,
        password=password,
        first_name=first_name,
        last_name=last_name,
        notes=notes,
        archive_size_quota_bytes=archive_size_quota,
    )


def _change_organization(sdk, username, org_id):
    user_id = _get_legacy_user_id(sdk, username)
    org_id = _get_org_id(sdk, org_id)
    return sdk.users.change_org_assignment(user_id=int(user_id), org_id=int(org_id))


def _get_org_id(sdk, org_id):
    org = sdk.orgs.get_by_uid(org_id)
    return org["orgId"]


def _deactivate_user(sdk, username):
    user_id = _get_legacy_user_id(sdk, username)
    sdk.users.deactivate(user_id)


def _reactivate_user(sdk, username):
    user_id = _get_legacy_user_id(sdk, username)
    sdk.users.reactivate(user_id)


def _get_user(sdk, username):
    # use when retrieving the user risk profile information
    try:
        return sdk.userriskprofile.get_by_username(username)
    except Py42UserRiskProfileNotFound:
        raise UserDoesNotExistError(username)


def _add_cloud_alias(sdk, username, alias):
    user = _get_user(sdk, username)
    sdk.userriskprofile.add_cloud_aliases(user["userId"], alias)


def _remove_cloud_alias(sdk, username, alias):
    user = _get_user(sdk, username)
    sdk.userriskprofile.delete_cloud_aliases(user["userId"], alias)


def _update_userriskprofile(
    sdk, append_notes=False, username=None, start_date=None, end_date=None, notes=None
):
    user = _get_user(sdk, username)
    user_id = user["userId"]
    if append_notes and notes != "null":
        notes = user["notes"] + f"\n\n{notes}"

    # py42 interprets empty string as "clear this value" for kwarg values. Since empty CSV columns
    # get parsed as "" we want to have user provide explicit 'null' string to indicate desire to
    # clear instead of just not update value
    start_date = (
        None if start_date == "" else ("" if start_date == "null" else start_date)
    )
    end_date = None if end_date == "" else ("" if end_date == "null" else end_date)
    notes = None if notes == "" else ("" if notes == "null" else notes)

    updated_user = sdk.userriskprofile.update(
        user_id, start_date=start_date, end_date=end_date, notes=notes
    )
    return {
        k: v
        for k, v in updated_user.data.items()
        if k in ["username", "userId", "startDate", "endDate", "notes"]
    }



import os
import signal
import site
import sys
import warnings

import click
from click_plugins import with_plugins
from pkg_resources import iter_entry_points
from py42.settings import set_user_agent_suffix

from code42cli import BANNER
from code42cli import PRODUCT_NAME
from code42cli.click_ext.groups import ExceptionHandlingGroup
from code42cli.cmds.alert_rules import alert_rules
from code42cli.cmds.alerts import alerts
from code42cli.cmds.auditlogs import audit_logs
from code42cli.cmds.cases import cases
from code42cli.cmds.departing_employee import departing_employee
from code42cli.cmds.devices import devices
from code42cli.cmds.high_risk_employee import high_risk_employee
from code42cli.cmds.legal_hold import legal_hold
from code42cli.cmds.profile import profile
from code42cli.cmds.securitydata import security_data
from code42cli.cmds.shell import shell
from code42cli.cmds.trustedactivities import trusted_activities
from code42cli.cmds.users import users
from code42cli.cmds.watchlists import watchlists
from code42cli.options import sdk_options

warnings.simplefilter("ignore", DeprecationWarning)


# Handle KeyboardInterrupts by just exiting instead of printing out a stack
def exit_on_interrupt(signal, frame):
    click.echo(err=True)
    sys.exit(1)


signal.signal(signal.SIGINT, exit_on_interrupt)

# Sets part of the user agent string that py42 attaches to requests for the purposes of
# identifying CLI users.
set_user_agent_suffix(PRODUCT_NAME)

CONTEXT_SETTINGS = {
    "help_option_names": ["-h", "--help"],
    "max_content_width": 200,
}


@with_plugins(iter_entry_points("code42cli.plugins"))
@click.group(
    cls=ExceptionHandlingGroup,
    context_settings=CONTEXT_SETTINGS,
    help=BANNER,
    invoke_without_command=True,
    no_args_is_help=True,
)
@click.option(
    "--python",
    is_flag=True,
    help="Print path to the python interpreter env that `code42cli` is installed in.",
)
@click.option(
    "--script-dir",
    is_flag=True,
    help="Print the directory the `code42` script was installed in (for adding to your PATH if needed).",
)
@sdk_options(hidden=True)
def cli(state, python, script_dir):
    if python:
        click.echo(sys.executable)
        sys.exit(0)
    if script_dir:
        for root, _dirs, files in os.walk(site.PREFIXES[0]):
            if "code42" in files or "code42.exe" in files:
                print(root)
                sys.exit(0)

        for root, _dirs, files in os.walk(site.USER_BASE):
            if "code42" in files or "code42.exe" in files:
                print(root)
                sys.exit(0)


cli.add_command(alerts)
cli.add_command(alert_rules)
cli.add_command(audit_logs)
cli.add_command(cases)
cli.add_command(departing_employee)
cli.add_command(devices)
cli.add_command(high_risk_employee)
cli.add_command(legal_hold)
cli.add_command(profile)
cli.add_command(security_data)
cli.add_command(shell)
cli.add_command(users)
cli.add_command(trusted_activities)
cli.add_command(watchlists)



from click import style

import code42cli.password as password
from code42cli.cmds.search.cursor_store import get_all_cursor_stores_for_profile
from code42cli.config import config_accessor
from code42cli.config import ConfigAccessor
from code42cli.config import NoConfigProfileError
from code42cli.errors import Code42CLIError


class Code42Profile:
    def __init__(self, profile):
        self._profile = profile

    @property
    def name(self):
        return self._profile.name

    @property
    def authority_url(self):
        return self._profile[ConfigAccessor.AUTHORITY_KEY]

    @property
    def username(self):
        return self._profile[ConfigAccessor.USERNAME_KEY]

    @property
    def ignore_ssl_errors(self):
        return self._profile[ConfigAccessor.IGNORE_SSL_ERRORS_KEY]

    @property
    def has_stored_password(self):
        stored_password = password.get_stored_password(self)
        return stored_password is not None and stored_password != ""

    def get_password(self):
        pwd = password.get_stored_password(self)
        if not pwd:
            pwd = password.get_password_from_prompt()
        return pwd

    def __str__(self):
        return (
            f"{self.name}: Username={self.username}, Authority URL={self.authority_url}"
        )


def _get_profile(profile_name=None):
    """Returns the profile for the given name."""
    config_profile = config_accessor.get_profile(profile_name)
    return Code42Profile(config_profile)


def get_profile(profile_name=None):
    if profile_name is None:
        validate_default_profile()
    try:
        return _get_profile(profile_name)
    except NoConfigProfileError as ex:
        raise Code42CLIError(str(ex), help=CREATE_PROFILE_HELP)


def default_profile_exists():
    try:
        profile = _get_profile()
        return profile.name and profile.name != ConfigAccessor.DEFAULT_VALUE
    except NoConfigProfileError:
        return False


def is_default_profile(name):
    if default_profile_exists():
        default = get_profile()
        return name == default.name


def validate_default_profile():
    if not default_profile_exists():
        existing_profiles = get_all_profiles()
        if not existing_profiles:
            raise Code42CLIError("No existing profile.", help=CREATE_PROFILE_HELP)
        else:
            raise Code42CLIError(
                "No default profile set.",
                help=_get_set_default_profile_help(existing_profiles),
            )


def profile_exists(profile_name=None):
    try:
        _get_profile(profile_name)
        return True
    except NoConfigProfileError:
        return False


def switch_default_profile(profile_name):
    profile = get_profile(profile_name)  # Handles if profile does not exist.
    config_accessor.switch_default_profile(profile.name)


def create_profile(name, server, username, ignore_ssl_errors):
    if profile_exists(name):
        raise Code42CLIError(f"A profile named '{name}' already exists.")
    config_accessor.create_profile(name, server, username, ignore_ssl_errors)


def delete_profile(profile_name):
    profile = _get_profile(profile_name)
    profile_name = profile.name
    if password.get_stored_password(profile) is not None:
        password.delete_password(profile)
    cursor_stores = get_all_cursor_stores_for_profile(profile_name)
    for store in cursor_stores:
        store.clean()
    config_accessor.delete_profile(profile_name)


def update_profile(name, server, username, ignore_ssl_errors):
    config_accessor.update_profile(name, server, username, ignore_ssl_errors)


def get_all_profiles():
    profiles = [
        Code42Profile(profile) for profile in config_accessor.get_all_profiles()
    ]
    return profiles


def get_stored_password(profile_name=None):
    profile = get_profile(profile_name)
    return password.get_stored_password(profile)


def set_password(new_password, profile_name=None):
    profile = get_profile(profile_name)
    password.set_password(profile, new_password)


CREATE_PROFILE_HELP = "\nTo add a profile, use:\n{}".format(
    style(
        "\tcode42 profile create "
        "--name <profile-name> "
        "--server <authority-URL> "
        "--username <username>\n",
        bold=True,
    )
)


def _get_set_default_profile_help(existing_profiles):
    existing_profiles = [str(profile) for profile in existing_profiles]
    help_msg = """
Use the --profile flag to specify which profile to use.

To set the default profile (used whenever --profile argument is not provided), use:
    {}

Existing profiles:
\t{}""".format(
        style("code42 profile use <profile-name>", bold=True),
        "\n\t".join(existing_profiles),
    )
    return help_msg



import json
import os
import shutil
from datetime import timezone
from functools import wraps
from hashlib import md5
from os import path
from signal import getsignal
from signal import SIGINT
from signal import signal

import dateutil.parser
from click import echo
from click import get_current_context
from click import style

_PADDING_SIZE = 3


def does_user_agree(prompt):
    """Prompts the user and checks if they said yes. If command has the `yes_option` flag, and
    `-y/--yes` is passed, this will always return `True`.
    """
    ctx = get_current_context()
    if ctx.obj.assume_yes:
        return True
    ans = input(prompt)
    ans = ans.strip().lower()
    return ans == "y"


def get_user_project_path(*subdirs):
    """The path on your user dir to /.code42cli/[subdir]."""
    package_name = __name__.split(".")[0]
    home = path.expanduser("~")
    hidden_package_name = f".{package_name}"
    user_project_path = path.join(home, hidden_package_name)
    result_path = path.join(user_project_path, *subdirs)
    if not path.exists(result_path):
        os.makedirs(result_path)

    return result_path


def find_format_width(records, header, include_header=True):
    """Fetches needed keys/items to be displayed based on header keys.

    Finds the largest string against each column so as to decide the padding size for the column.

    Args:
        records (list or dict): A list of data to be formatted.
        header (dict): Key-value where keys should map to keys of record dict and
          value is the corresponding column name to be displayed on the CLI.
        include_header (bool): Include header in output, defaults to True.

    Returns:
        tuple (list of dict, dict): i.e Filtered records, padding size of columns.
    """
    if isinstance(records, dict):
        records = [records]

    rows = []
    if include_header:
        if not header:
            header = _get_default_header(records)
        rows.append(header)
    widths = dict(header.items())  # Copy
    for record_row in records:
        row = {}
        for header_key in header.keys():
            item = record_row.get(header_key)
            row[header_key] = item
            widths[header_key] = max(widths[header_key], str(item), key=len)
        rows.append(row)
    column_sizes = {key: len(value) for key, value in widths.items()}
    return rows, column_sizes


def format_to_table(rows, column_size):
    """Formats given rows into a string of left justified table."""
    lines = []
    for row in rows:
        line = ""
        for key in row.keys():
            line += str(row[key]).ljust(column_size[key] + _PADDING_SIZE)
        lines.append(line)
    return "\n".join(lines)


def format_string_list_to_columns(string_list, max_width=None):
    """Prints a list of strings in justified columns and fits them neatly into specified width."""
    if not string_list:
        return
    if not max_width:
        max_width, _ = shutil.get_terminal_size()
    column_width = len(max(string_list, key=len)) + _PADDING_SIZE
    num_columns = int(max_width / column_width) or 1
    format_string = f"{{:<{column_width}}}" * num_columns
    batches = [
        string_list[i : i + num_columns]
        for i in range(0, len(string_list), num_columns)
    ]
    padding = ["" for _ in range(num_columns)]
    for batch in batches:
        echo(format_string.format(*batch + padding))
    echo()


class warn_interrupt:
    """A context decorator class used to wrap functions where a keyboard interrupt could potentially
    leave things in a bad state. Warns the user with provided message and exits when wrapped
    function is complete. Requires user to ctrl-c a second time to force exit.

    Usage:

    @warn_interrupt(warning="example message")
    def my_important_func():
        pass
    """

    def __init__(self, warning="Cancelling operation cleanly, one moment... "):
        self.warning = warning
        self.old_int_handler = None
        self.interrupted = False
        self.exit_instructions = style("Hit CTRL-C again to force quit.", fg="red")

    def __enter__(self):
        self.old_int_handler = getsignal(SIGINT)
        signal(SIGINT, self._handle_interrupts)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.interrupted:
            exit(1)
        signal(SIGINT, self.old_int_handler)

        return False

    def _handle_interrupts(self, sig, frame):
        if not self.interrupted:
            self.interrupted = True
            echo(f"\n{self.warning}\n{self.exit_instructions}", err=True)
        else:
            exit()

    def __call__(self, func):
        @wraps(func)
        def inner(*args, **kwargs):
            with self:
                return func(*args, **kwargs)

        return inner


def get_url_parts(url_str):
    parts = url_str.split(":")
    port = None
    if len(parts) > 1 and parts[1] != "":
        port = int(parts[1])

    return parts[0], port


def _get_default_header(header_items):
    if not header_items:
        return

    # Creates dict where keys and values are the same for `find_format_width()`.
    header = {}
    for item in header_items:
        keys = item.keys()
        for key in keys:
            if key not in header and isinstance(key, str):
                header[key] = key

    return header


def hash_event(event):
    if isinstance(event, dict):
        event = json.dumps(event, sort_keys=True)
    return md5(event.encode()).hexdigest()


def print_numbered_list(items):
    """Outputs a numbered list of items to the user.
    For example, provide ["test", "foo"] to print "1. test\n2. foo".
    """

    choices = dict(enumerate(items, 1))
    for num in choices:
        echo(f"{num}. {choices[num]}")
    echo()


def parse_timestamp(date_str):
    # example: {"property": "bar", "timestamp": "2020-11-23T17:13:26.239647Z"}
    ts = date_str[:-1]
    date = dateutil.parser.parse(ts).replace(tzinfo=timezone.utc)
    return date.timestamp()


def deprecation_warning(text):
    echo(style(text, fg="red"), err=True)



import click

username_arg = click.argument("username")
cloud_alias_option = click.option(
    "--cloud-alias",
    help="If the employee has an email alias other than their Code42 username "
    "that they use for cloud services such as Google Drive, OneDrive, or Box, "
    "add and monitor the alias. WARNING: Adding a cloud alias will override any "
    "existing cloud alias for this user.",
)
notes_option = click.option("--notes", help="Optional notes about the employee.")



import csv

import click

from code42cli.click_ext.types import AutoDecodedFile
from code42cli.errors import Code42CLIError


def read_csv_arg(headers):
    """Helper for defining arguments that read from a csv file. Automatically converts
    the file name provided on command line to a list of csv rows (passed to command
    function as `csv_rows` param).
    """
    return click.argument(
        "csv_rows",
        metavar="CSV_FILE",
        type=AutoDecodedFile("r"),
        callback=lambda ctx, param, arg: read_csv(arg, headers=headers),
    )


def read_csv(file, headers):
    """Helper to read a csv file object into a list of dict rows.
    If CSV has a header row, all items in `headers` arg must be present in CSV or an
    error is raised. Any extra columns will get filtered out from resulting dicts.

    If no header row is present in CSV, column count must match `headers` arg length or
    else error is raised.
    """
    lines = file.readlines()

    # check if header is commented for flat-file backwards compatability
    if lines[0].startswith("#"):
        # strip comment line
        lines.pop(0)

    first_line = lines[0].strip().split(",")

    # handle when first row has all of our expected headers
    if all(field in first_line for field in headers):
        reader = csv.DictReader(lines[1:], fieldnames=first_line)
        csv_rows = [{key: row[key] for key in headers} for row in reader]
        if not csv_rows:
            raise Code42CLIError("CSV contains no data rows.")
        return csv_rows

    # handle when first row has no expected headers
    elif all(field not in first_line for field in headers):
        #  only process header-less CSVs if we get exact expected column count
        if len(first_line) == len(headers):
            return list(csv.DictReader(lines, fieldnames=headers))
        else:
            raise Code42CLIError(
                "CSV data is ambiguous. Column count must match expected columns exactly when no "
                f"header row is present. Expected columns: {headers}"
            )
    # handle when first row has some expected headers but not all
    else:
        missing = [field for field in headers if field not in first_line]
        raise Code42CLIError(f"Missing required columns in csv: {missing}")



import csv
import io
import json
from itertools import chain
from typing import Generator

import click
from pandas import concat
from pandas import notnull

from code42cli.enums import FileEventsOutputFormat
from code42cli.enums import OutputFormat
from code42cli.errors import Code42CLIError
from code42cli.logger.formatters import CEF_TEMPLATE
from code42cli.logger.formatters import map_event_to_cef
from code42cli.util import find_format_width
from code42cli.util import format_to_table


CEF_DEFAULT_PRODUCT_NAME = "Advanced Exfiltration Detection"
CEF_DEFAULT_SEVERITY_LEVEL = "5"

# Uses method `echo_via_pager()` when 10 or more records.
OUTPUT_VIA_PAGER_THRESHOLD = 10


class OutputFormatter:
    def __init__(self, output_format, header=None):
        output_format = output_format.upper() if output_format else OutputFormat.TABLE
        self.output_format = output_format
        self._format_func = to_table
        self.header = header

        if output_format == OutputFormat.CSV:
            self._format_func = to_csv
        elif output_format == OutputFormat.RAW:
            self._format_func = to_json
        elif output_format == OutputFormat.TABLE:
            self._format_func = self._to_table
        elif output_format == OutputFormat.JSON:
            self._format_func = to_formatted_json

    def _format_output(self, output, *args, **kwargs):
        return self._format_func(output, *args, **kwargs)

    def _to_table(self, output, include_header=True):
        return to_table(output, self.header, include_header=include_header)

    def get_formatted_output(self, output):
        if self._requires_list_output:
            yield self._format_output(output)
        else:
            for item in output:
                yield self._format_output(item)

    def echo_formatted_list(self, output_list, force_pager=False):
        formatted_output = self.get_formatted_output(output_list)
        if len(output_list) > OUTPUT_VIA_PAGER_THRESHOLD or force_pager:
            click.echo_via_pager(formatted_output)
        else:
            for output in formatted_output:
                click.echo(output, nl=False)
            if self.output_format in [OutputFormat.TABLE]:
                click.echo()

    @property
    def _requires_list_output(self):
        return self.output_format in (OutputFormat.TABLE, OutputFormat.CSV)


class DataFrameOutputFormatter:
    def __init__(self, output_format, checkpoint_func=None):
        self.output_format = (
            output_format.upper() if output_format else OutputFormat.TABLE
        )
        if self.output_format not in OutputFormat.choices():
            raise Code42CLIError(
                f"DataFrameOutputFormatter received an invalid format: {self.output_format}"
            )
        self.checkpoint_func = checkpoint_func or (lambda x: None)

    def _ensure_iterable(self, dfs):
        if not isinstance(dfs, (Generator, list, tuple)):
            return [dfs]
        return dfs

    def _iter_table(self, dfs, columns=None, **kwargs):
        dfs = self._ensure_iterable(dfs)
        df = concat(dfs)
        if df.empty:
            return
        # convert everything to strings so we can left-justify format
        df = df.fillna("").applymap(str)
        # set overrideable default kwargs
        kwargs = {
            "index": False,
            "justify": "left",
            "formatters": make_left_aligned_formatter(df),
            **kwargs,
        }
        if columns:
            filtered = self._select_columns(df, columns)
            formatted_rows = filtered.to_string(**kwargs).splitlines(keepends=True)
        else:
            formatted_rows = df.to_string(**kwargs).splitlines(keepends=True)
        # don't checkpoint the header row
        if kwargs.get("header") is not False:
            yield formatted_rows.pop(0)

        yield from self._checkpoint_and_iter_formatted_events(df, formatted_rows)

    def _iter_csv(self, dfs, columns=None, **kwargs):
        dfs = self._ensure_iterable(dfs)
        no_header = kwargs.get("header") is False

        for i, df in enumerate(dfs):
            if df.empty:
                continue
            # convert null values to empty string
            df.fillna("", inplace=True)
            # only add header on first df and if header=False was not passed in kwargs
            header = False if no_header else (i == 0)
            kwargs = {"index": False, "header": header, **kwargs}
            if columns:
                filtered = self._select_columns(df, columns)
                formatted_rows = filtered.to_csv(**kwargs).splitlines(keepends=True)
            else:
                formatted_rows = df.to_csv(**kwargs).splitlines(keepends=True)
            if header:
                yield formatted_rows.pop(0)

            yield from self._checkpoint_and_iter_formatted_events(df, formatted_rows)

    def _iter_json(self, dfs, columns=None, **kwargs):
        kwargs = {"ensure_ascii": False, **kwargs}
        for event in self.iter_rows(dfs, columns=columns):
            json_string = json.dumps(event, **kwargs)
            yield f"{json_string}\n"

    def _checkpoint_and_iter_formatted_events(self, df, formatted_rows):
        for event, row in zip(df.to_dict("records"), formatted_rows):
            yield row
            self.checkpoint_func(event)

    def _echo_via_pager_if_over_threshold(self, gen):
        first_rows = []
        try:
            for _ in range(OUTPUT_VIA_PAGER_THRESHOLD):
                first_rows.append(next(gen))
        except StopIteration:
            click.echo("".join(first_rows))
            return

        click.echo_via_pager(chain(first_rows, gen))

    def _select_columns(self, df, columns):
        if df.empty:
            return df
        if not isinstance(columns, (list, tuple)):
            raise Code42CLIError(
                "'columns' parameter must be a list or tuple of column names."
            )
        # enable case-insensitive column selection
        normalized_map = {c.lower(): c for c in df.columns}
        try:
            columns = [normalized_map[c.lower()] for c in columns]
            return df[columns]
        except KeyError as e:
            key = e.args[0]
            raise click.BadArgumentUsage(
                f"'{key}' is not a valid column. Valid columns are: {list(df.columns)}"
            )

    def iter_rows(self, dfs, columns=None):
        """
        Accepts a pandas DataFrame or list/generator of DataFrames and yields each
        'row' of the DataFrame as a dict, calling the `checkpoint_func` on each row
        after it has been yielded.

        Accepts an optional list of column names that filter
        columns in the yielded results.
        """
        dfs = self._ensure_iterable(dfs)
        for df in dfs:
            # convert pandas' default null (numpy.NaN) to None
            df = df.astype(object).where(notnull, None)
            if columns:
                filtered = self._select_columns(df, columns)
            else:
                filtered = df
            for full_event, filtered_event in zip(
                df.to_dict("records"), filtered.to_dict("records")
            ):
                yield filtered_event
                self.checkpoint_func(full_event)

    def get_formatted_output(self, dfs, columns=None, **kwargs):
        """
        Accepts a pandas DataFrame or list/generator of DataFrames and formats and yields
        the results line by line to the caller as a generator.

        Accepts an optional list of column names that filter columns in the yielded
        results.

        Any additional kwargs provided will be passed to the underlying format method
        if customizations are required.
        """
        if self.output_format == OutputFormat.TABLE:
            yield from self._iter_table(dfs, columns=columns, **kwargs)

        elif self.output_format == OutputFormat.CSV:
            yield from self._iter_csv(dfs, columns=columns, **kwargs)

        elif self.output_format == OutputFormat.JSON:
            kwargs = {"indent": 4, **kwargs}
            yield from self._iter_json(dfs, columns=columns, **kwargs)

        elif self.output_format == OutputFormat.RAW:
            yield from self._iter_json(dfs, columns=columns, **kwargs)

        else:
            raise Code42CLIError(
                f"DataFrameOutputFormatter received an invalid format: {self.output_format}"
            )

    def echo_formatted_dataframes(
        self, dfs, columns=None, force_pager=False, force_no_pager=False, **kwargs
    ):
        """
        Accepts a pandas DataFrame or list/generator of DataFrames and formats and echos the
        result to stdout. If total lines > 10, results will be sent to pager. `force_pager`
        and `force_no_pager` can be set to override the pager logic based on line count.

        Accepts an optional list of column names that filter
        columns in the echoed results.

        Any additional kwargs provided will be passed to the underlying format method
        if customizations are required.
        """
        lines = self.get_formatted_output(dfs, columns=columns, **kwargs)
        try:
            # check for empty generator
            first = next(lines)
            lines = chain([first], lines)
        except StopIteration:
            click.echo("No results found.")
            return
        if force_pager and force_no_pager:
            raise Code42CLIError("force_pager cannot be used with force_no_pager.")
        if force_pager:
            click.echo_via_pager(lines)
        elif force_no_pager:
            for line in lines:
                click.echo(line)
        else:
            self._echo_via_pager_if_over_threshold(lines)


class FileEventsOutputFormatter(DataFrameOutputFormatter):
    """Class that adds CEF format output option to base DataFrameOutputFormatter."""

    def __init__(self, output_format, checkpoint_func=None):
        self.output_format = (
            output_format.upper() if output_format else OutputFormat.RAW
        )
        if self.output_format not in FileEventsOutputFormat.choices():
            raise Code42CLIError(
                f"FileEventsOutputFormatter received an invalid format: {self.output_format}"
            )
        self.checkpoint_func = checkpoint_func or (lambda x: None)

    def _iter_cef(self, dfs, **kwargs):
        dfs = self._ensure_iterable(dfs)
        for df in dfs:
            df = df.mask(df.isna(), other=None)
            for _i, row in df.iterrows():
                event = dict(row)
                yield f"{_convert_event_to_cef(event)}\n"
                self.checkpoint_func(event)

    def get_formatted_output(self, dfs, columns=None, **kwargs):
        if self.output_format == FileEventsOutputFormat.CEF:
            yield from self._iter_cef(dfs, **kwargs)
        else:
            yield from super().get_formatted_output(dfs, columns=columns, **kwargs)


def to_csv(output):
    """Output is a list of records"""

    if not output:
        return
    string_io = io.StringIO(newline=None)
    fieldnames = list({k for d in output for k in d.keys()})
    writer = csv.DictWriter(string_io, fieldnames=fieldnames)
    writer.writeheader()
    writer.writerows(output)
    return string_io.getvalue()


def to_table(output, header, include_header=True):
    """Output is a list of records"""
    if not output:
        return

    rows, column_size = find_format_width(output, header, include_header=include_header)
    return format_to_table(rows, column_size)


def to_json(output):
    """Output is a single record"""
    return f"{json.dumps(output)}\n"


def to_formatted_json(output):
    """Output is a single record"""
    return f"{json.dumps(output, indent=4)}\n"


def to_cef(output):
    """Output is a single record"""
    return f"{_convert_event_to_cef(output)}\n"


def _convert_event_to_cef(event):
    ext, evt, sig_id = map_event_to_cef(event)
    cef_log = CEF_TEMPLATE.format(
        productName=CEF_DEFAULT_PRODUCT_NAME,
        signatureID=sig_id,
        eventName=evt,
        severity=CEF_DEFAULT_SEVERITY_LEVEL,
        extension=ext,
    )
    return cef_log


def make_left_aligned_formatter(df):
    return {c: f"{{:<{df[c].str.len().max()}s}}".format for c in df.columns}



from collections import OrderedDict

import click
from click import echo
from py42.exceptions import Py42BadRequestError
from py42.util import format_json

from code42cli import PRODUCT_NAME
from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.cmds.shared import get_user_id
from code42cli.errors import Code42CLIError
from code42cli.file_readers import read_csv_arg
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.output_formats import OutputFormatter


class AlertRuleTypes:
    EXFILTRATION = "FED_ENDPOINT_EXFILTRATION"
    CLOUD_SHARE = "FED_CLOUD_SHARE_PERMISSIONS"
    FILE_TYPE_MISMATCH = "FED_FILE_TYPE_MISMATCH"


_HEADER_KEYS_MAP = OrderedDict()
_HEADER_KEYS_MAP["observerRuleId"] = "RuleId"
_HEADER_KEYS_MAP["name"] = "Name"
_HEADER_KEYS_MAP["severity"] = "Severity"
_HEADER_KEYS_MAP["type"] = "Type"
_HEADER_KEYS_MAP["ruleSource"] = "Source"
_HEADER_KEYS_MAP["isEnabled"] = "Enabled"


@click.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def alert_rules(state):
    """Manage users associated with alert rules."""
    pass


rule_id_option = click.option(
    "--rule-id", required=True, help="Identification number of the alert rule."
)


@alert_rules.command()
@rule_id_option
@click.option(
    "-u",
    "--username",
    required=True,
    help="The username of the user to add to the alert rule.",
)
@sdk_options()
def add_user(state, rule_id, username):
    """Add a user to an alert rule."""
    _add_user(state.sdk, rule_id, username)


@alert_rules.command()
@rule_id_option
@click.option(
    "-u",
    "--username",
    required=True,
    help="The username of the user to remove from the alert rule.",
)
@sdk_options()
def remove_user(state, rule_id, username):
    """Remove a user from an alert rule."""
    try:
        _remove_user(state.sdk, rule_id, username)
    except Py42BadRequestError:
        raise Code42CLIError(
            f"User {username} is not currently assigned to rule-id {rule_id}."
        )


@alert_rules.command("list")
@format_option
@sdk_options()
def list_alert_rules(state, format):
    """Fetch existing alert rules."""
    formatter = OutputFormatter(format, _HEADER_KEYS_MAP)
    selected_rules = _get_all_rules_metadata(state.sdk)

    if selected_rules:
        formatter.echo_formatted_list(selected_rules)


@alert_rules.command()
@click.argument("rule_id")
@sdk_options()
def show(state, rule_id):
    """Print out detailed alert rule criteria."""
    selected_rule = _get_rule_metadata(state.sdk, rule_id)
    if selected_rule:
        get = _get_rule_type_func(state.sdk, selected_rule[0]["type"])
        rule_detail = get(rule_id)
        echo(format_json(rule_detail.text))


@alert_rules.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def bulk(state):
    """Tools for executing bulk alert rule actions."""
    pass


ALERT_RULES_CSV_HEADERS = ["rule_id", "username"]

alert_rules_generate_template = generate_template_cmd_factory(
    group_name="alert_rules",
    commands_dict={"add": ALERT_RULES_CSV_HEADERS, "remove": ALERT_RULES_CSV_HEADERS},
)
bulk.add_command(alert_rules_generate_template)


@bulk.command(
    help=f"Bulk add users to alert rules from a CSV file. "
    f"CSV file format: {','.join(ALERT_RULES_CSV_HEADERS)}"
)
@read_csv_arg(headers=ALERT_RULES_CSV_HEADERS)
@sdk_options()
def add(state, csv_rows):
    sdk = state.sdk

    def handle_row(rule_id, username):
        _add_user(sdk, rule_id, username)

    run_bulk_process(
        handle_row, csv_rows, progress_label="Adding users to alert-rules:"
    )


@bulk.command(
    help="Bulk remove users from alert rules using a CSV file. "
    "CSV file format: {','.join(ALERT_RULES_CSV_HEADERS)}"
)
@read_csv_arg(headers=ALERT_RULES_CSV_HEADERS)
@sdk_options()
def remove(state, csv_rows):
    sdk = state.sdk

    def handle_row(rule_id, username):
        _remove_user(sdk, rule_id, username)

    run_bulk_process(
        handle_row, csv_rows, progress_label="Removing users from alert-rules:"
    )


def _add_user(sdk, rule_id, username):
    user_id = get_user_id(sdk, username)
    rules = _get_rule_metadata(sdk, rule_id)
    if rules:
        sdk.alerts.rules.add_user(rule_id, user_id)


def _remove_user(sdk, rule_id, username):
    user_id = get_user_id(sdk, username)
    rules = _get_rule_metadata(sdk, rule_id)
    if rules:
        sdk.alerts.rules.remove_user(rule_id, user_id)


def _get_all_rules_metadata(sdk):
    rules_generator = sdk.alerts.rules.get_all()
    selected_rules = [
        rule for rules in rules_generator for rule in rules["ruleMetadata"]
    ]
    return _handle_rules_results(selected_rules)


def _get_rule_metadata(sdk, rule_id):
    rules = sdk.alerts.rules.get_by_observer_id(rule_id)["ruleMetadata"]
    return _handle_rules_results(rules, rule_id)


def _handle_rules_results(rules, rule_id=None):
    if not rules:
        id_msg = f"with RuleId {rule_id} " if rule_id else ""
        msg = f"No alert rules {id_msg}found."
        raise Code42CLIError(msg)
    return rules


def _get_rule_type_func(sdk, rule_type):
    if rule_type == AlertRuleTypes.EXFILTRATION:
        return sdk.alerts.rules.exfiltration.get
    elif rule_type == AlertRuleTypes.CLOUD_SHARE:
        return sdk.alerts.rules.cloudshare.get
    elif rule_type == AlertRuleTypes.FILE_TYPE_MISMATCH:
        return sdk.alerts.rules.filetypemismatch.get
    else:
        raise Code42CLIError(
            "Received an unknown rule type from server. You might need to update "
            f"to a newer version of {PRODUCT_NAME}"
        )



from codecs import open
from os import path

from setuptools import find_packages
from setuptools import setup

here = path.abspath(path.dirname(__file__))

about = {}
with open(path.join(here, "src", "code42cli", "__version__.py"), encoding="utf8") as fh:
    exec(fh.read(), about)

with open(path.join(here, "README.md"), "r", "utf-8") as f:
    readme = f.read()

setup(
    name="code42cli",
    version=about["__version__"],
    url="https://github.com/code42/py42",
    project_urls={
        "Issue Tracker": "https://github.com/code42/code42cli/issues",
        "Documentation": "https://clidocs.code42.com/",
        "Source Code": "https://github.com/code42/code42cli",
    },
    description="The official command line tool for interacting with Code42",
    long_description=readme,
    long_description_content_type="text/markdown",
    packages=find_packages("src"),
    package_dir={"": "src"},
    include_package_data=True,
    zip_safe=False,
    python_requires=">=3.6.2, <4",
    install_requires=[
        "chardet",
        "click>=7.1.1, <8",
        "click_plugins>=1.1.1",
        "colorama>=0.4.3",
        "keyring==18.0.1",
        "keyrings.alt==3.2.0",
        "ipython==7.16.3",
        "pandas>=1.1.3",
        "py42>=1.23.0",
    ],
    extras_require={
        "dev": [
            "flake8==3.8.3",
            "pytest==4.6.11",
            "pytest-cov==2.10.0",
            "pytest-mock==2.0.0",
            "tox>=3.17.1",
        ],
        "docs": [
            "sphinx==4.4.0",
            "myst-parser==0.16",
            "sphinx_rtd_theme==1.0.0",
            "sphinx-click",
        ],
    },
    classifiers=[
        "Intended Audience :: Developers",
        "Natural Language :: English",
        "License :: OSI Approved :: MIT License",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.6",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: Implementation :: CPython",
    ],
    entry_points={"console_scripts": ["code42=code42cli.main:cli"]},
)



import json
from datetime import datetime

import click
from py42.sdk.queries.query_filter import FilterGroup

from code42cli.click_ext.options import incompatible_with
from code42cli.click_ext.types import FileOrString
from code42cli.enums import SendToFileEventsOutputFormat
from code42cli.logger.enums import ServerProtocol


def is_in_filter(filter_cls):
    def callback(ctx, param, arg):
        if arg:
            ctx.obj.search_filters.append(filter_cls.is_in(arg))
        return arg

    return callback


def not_in_filter(filter_cls):
    def callback(ctx, param, arg):
        if arg:
            ctx.obj.search_filters.append(filter_cls.not_in(arg))
        return arg

    return callback


def exists_filter(filter_cls):
    def callback(ctx, param, arg):
        if not arg:
            ctx.obj.search_filters.append(filter_cls.exists())
            return arg

    return callback


def contains_filter(filter_cls):
    def callback(ctx, param, arg):
        if arg:
            for item in arg:
                ctx.obj.search_filters.append(filter_cls.contains(item))
        return arg

    return callback


def not_contains_filter(filter_cls):
    def callback(ctx, param, arg):
        if arg:
            for item in arg:
                ctx.obj.search_filters.append(filter_cls.not_contains(item))
        return arg

    return callback


AdvancedQueryAndSavedSearchIncompatible = incompatible_with(
    ["advanced_query", "saved_search"]
)


class BeginOption(AdvancedQueryAndSavedSearchIncompatible):
    """click.Option subclass that enforces correct --begin option usage."""

    def __init__(self, *args, **kwargs):
        super().__init__(*args, **kwargs)

    def handle_parse_result(self, ctx, opts, args):
        # if ctx.obj is None it means we're in autocomplete mode and don't want to validate
        if (
            ctx.obj is not None
            and "saved_search" not in opts
            and "advanced_query" not in opts
        ):
            profile = opts.get("profile") or ctx.obj.profile.name
            cursor = ctx.obj.cursor_getter(profile)
            checkpoint_arg_present = "use_checkpoint" in opts
            checkpoint_value = (
                cursor.get(opts.get("use_checkpoint", ""))
                if checkpoint_arg_present
                else None
            )
            begin_present = "begin" in opts
            if (
                checkpoint_arg_present
                and checkpoint_value is not None
                and begin_present
            ):
                opts.pop("begin")
                try:
                    checkpoint_value = datetime.utcfromtimestamp(
                        float(checkpoint_value)
                    )
                except ValueError:
                    pass
                click.echo(
                    "Ignoring --begin value as --use-checkpoint was passed and checkpoint of "
                    f"{checkpoint_value} exists.\n",
                    err=True,
                )
            if (
                checkpoint_arg_present
                and checkpoint_value is None
                and not begin_present
            ):
                raise click.UsageError(
                    message="--begin date is required for --use-checkpoint when no checkpoint "
                    "exists yet.",
                )
            if not checkpoint_arg_present and not begin_present:
                raise click.UsageError(message="--begin date is required.")
        return super().handle_parse_result(ctx, opts, args)


def _parse_query_from_json(ctx, param, arg):
    if arg is None:
        return
    try:
        query = json.loads(arg)
        filter_groups = [FilterGroup.from_dict(group) for group in query["groups"]]
        return filter_groups
    except json.JSONDecodeError as json_error:
        raise click.BadParameter(f"Unable to parse JSON: {json_error}")
    except KeyError as key_error:
        raise click.BadParameter(f"Unable to build query from input JSON: {key_error}")


def advanced_query_option(term, **kwargs):
    defaults = dict(
        help=f"A raw JSON {term} query. "
        "Useful for when the provided query parameters do not satisfy your requirements. "
        "Argument can be passed as a string, read from stdin by passing '-', or from a filename if "
        "prefixed with '@', e.g. '--advanced-query @query.json'. "
        "WARNING: Using advanced queries is incompatible with other query-building arguments.",
        metavar="QUERY_JSON",
        type=FileOrString(),
        callback=_parse_query_from_json,
    )
    defaults.update(kwargs)
    return click.option("--advanced-query", **defaults)


or_query_option = click.option(
    "--or-query",
    is_flag=True,
    cls=AdvancedQueryAndSavedSearchIncompatible,
    help="Combine query filter options with 'OR' logic instead of the default 'AND'.",
)

include_all_option = click.option(
    "--include-all",
    default=False,
    is_flag=True,
    help="Display simple properties of the primary level of the nested response.",
    cls=incompatible_with("columns"),
)


def server_options(f):
    hostname_arg = click.argument("hostname")
    protocol_option = click.option(
        "-p",
        "--protocol",
        type=click.Choice(ServerProtocol(), case_sensitive=False),
        default=ServerProtocol.UDP,
        help="Protocol used to send logs to server. "
        "Use TCP-TLS for additional security. Defaults to UDP.",
    )
    certs_option = click.option(
        "--certs",
        type=str,
        help="A CA certificates-chain file for the TCP-TLS protocol.",
    )
    ignore_cert_validation = click.option(
        "--ignore-cert-validation",
        help="Set to skip CA certificate validation. "
        "Incompatible with the 'certs' option.",
        is_flag=True,
        default=None,
        cls=incompatible_with(["certs"]),
    )
    f = hostname_arg(f)
    f = protocol_option(f)
    f = certs_option(f)
    f = ignore_cert_validation(f)
    return f


send_to_format_options = click.option(
    "-f",
    "--format",
    type=click.Choice(SendToFileEventsOutputFormat(), case_sensitive=False),
    help="The output format of the result. Defaults to RAW-JSON format.",
    default=SendToFileEventsOutputFormat.RAW,
)



import json
import os

import click
from py42.clients.cases import CaseStatus
from py42.exceptions import Py42BadRequestError
from py42.exceptions import Py42CaseAlreadyHasEventError
from py42.exceptions import Py42NotFoundError
from py42.exceptions import Py42UpdateClosedCaseError

from code42cli.bulk import generate_template_cmd_factory
from code42cli.bulk import run_bulk_process
from code42cli.click_ext.groups import OrderedGroup
from code42cli.errors import Code42CLIError
from code42cli.file_readers import read_csv_arg
from code42cli.options import format_option
from code42cli.options import sdk_options
from code42cli.options import set_begin_default_dict
from code42cli.options import set_end_default_dict
from code42cli.output_formats import OutputFormatter


case_number_arg = click.argument("case-number", type=int)
case_number_option = click.option(
    "--case-number", type=int, help="The number assigned to the case.", required=True
)
name_option = click.option(
    "--name",
    help="The name of the case.",
)
assignee_option = click.option(
    "--assignee", help="The UID of the user to assign to the case."
)
description_option = click.option("--description", help="The description of the case.")
findings_option = click.option("--findings", help="Any findings for the case.")
subject_option = click.option(
    "--subject", help="The user UID of the subject of the case."
)
status_option = click.option(
    "--status",
    help="Status of the case. `OPEN` or `CLOSED`.",
    type=click.Choice(CaseStatus.choices()),
)
file_event_id_option = click.option(
    "--event-id", required=True, help="The file event ID associated with the case."
)
CASES_KEYWORD = "cases"
BEGIN_DATE_DICT = set_begin_default_dict(CASES_KEYWORD)
END_DATE_DICT = set_end_default_dict(CASES_KEYWORD)


def _get_cases_header():
    return {
        "number": "Number",
        "name": "Name",
        "assignee": "Assignee",
        "status": "Status",
        "subject": "Subject",
        "createdAt": "Creation Time",
        "updatedAt": "Last Update Time",
    }


def _get_events_header():
    return {
        "eventId": "Event Id",
        "eventTimestamp": "Timestamp",
        "filePath": "Path",
        "fileName": "File",
        "exposure": "Exposure",
    }


@click.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def cases(state):
    """Manage cases and events associated with cases."""
    pass


@cases.command()
@click.argument("name")
@assignee_option
@description_option
@findings_option
@subject_option
@sdk_options()
def create(state, name, subject, assignee, description, findings):
    """Create a new case."""
    state.sdk.cases.create(
        name,
        subject=subject,
        assignee=assignee,
        description=description,
        findings=findings,
    )


@cases.command()
@case_number_arg
@name_option
@assignee_option
@description_option
@findings_option
@subject_option
@status_option
@sdk_options()
def update(state, case_number, name, subject, assignee, description, findings, status):
    """Update case details for the given case."""
    state.sdk.cases.update(
        case_number,
        name=name,
        subject=subject,
        assignee=assignee,
        description=description,
        findings=findings,
        status=status,
    )


@cases.command("list")
@click.option(
    "--name",
    help="Filter by name of a case. Supports partial name matches.",
)
@click.option("--subject", help="Filter by the user UID of the subject of a case.")
@click.option("--assignee", help="Filter by the user UID of an assignee.")
@click.option("--begin-create-time", **BEGIN_DATE_DICT)
@click.option("--end-create-time", **END_DATE_DICT)
@click.option("--begin-update-time", **BEGIN_DATE_DICT)
@click.option("--end-update-time", **END_DATE_DICT)
@click.option("--status", help="Filter cases by case status.")
@format_option
@sdk_options()
def _list(
    state,
    name,
    assignee,
    subject,
    begin_create_time,
    end_create_time,
    begin_update_time,
    end_update_time,
    status,
    format,
):
    """List all the cases."""
    pages = state.sdk.cases.get_all(
        name=name,
        assignee=assignee,
        subject=subject,
        min_create_time=begin_create_time,
        max_create_time=end_create_time,
        min_update_time=begin_update_time,
        max_update_time=end_update_time,
        status=status,
    )
    formatter = OutputFormatter(format, _get_cases_header())
    cases = [case for page in pages for case in page["cases"]]
    if cases:
        formatter.echo_formatted_list(cases)
    else:
        click.echo("No cases found.")


def _get_file_events(sdk, case_number):
    response = sdk.cases.file_events.get_all(case_number)
    if not response["events"]:
        return None
    return json.loads(response.text)


def _display_file_events(events):
    if events:
        click.echo("\nFile Events:\n")
        click.echo(json.dumps(events, indent=4))
    else:
        click.echo("\nNo events found.")


@cases.command()
@case_number_arg
@click.option(
    "--include-file-events",
    is_flag=True,
    help="View file events associated to the case.",
)
@sdk_options()
@format_option
def show(state, case_number, format, include_file_events):
    """Show case details."""
    formatter = OutputFormatter(format)
    try:
        response = state.sdk.cases.get(case_number)
        formatter.echo_formatted_list([response.data])
        if include_file_events:
            events = _get_file_events(state.sdk, case_number)
            _display_file_events(events)
    except Py42NotFoundError:
        raise Code42CLIError(f"Invalid case-number {case_number}.")


@cases.command()
@case_number_arg
@click.option(
    "--path",
    help="The file path where to save the PDF. Defaults to the current directory.",
    default=os.getcwd(),
)
@sdk_options()
def export(state, case_number, path):
    """Download a case detail summary as a PDF file at the given path with name <case_number>_case_summary.pdf."""
    response = state.sdk.cases.export_summary(case_number)
    file = os.path.join(path, f"{case_number}_case_summary.pdf")
    with open(file, "wb") as f:
        f.write(response.content)


@cases.group(cls=OrderedGroup)
@sdk_options()
def file_events(state):
    """Fetch file events associated with the case."""
    pass


@file_events.command("list")
@case_number_arg
@sdk_options()
@format_option
def file_events_list(state, case_number, format):
    """List all the file events associated with the case."""
    formatter = OutputFormatter(format, _get_events_header())
    try:
        response = state.sdk.cases.file_events.get_all(case_number)
    except Py42NotFoundError:
        raise Code42CLIError("Invalid case-number.")

    if not response["events"]:
        click.echo("No events found.")
    else:
        events = [event for event in response["events"]]
        formatter.echo_formatted_list(events)


@file_events.command()
@case_number_option
@file_event_id_option
@sdk_options()
def add(state, case_number, event_id):
    """Associate a file event to a case, by event ID."""
    try:
        state.sdk.cases.file_events.add(case_number, event_id)
    except Py42UpdateClosedCaseError:
        raise
    except Py42CaseAlreadyHasEventError:
        raise
    except Py42BadRequestError:
        raise Code42CLIError("Invalid case-number or event-id.")


@file_events.command()
@case_number_option
@file_event_id_option
@sdk_options()
def remove(state, case_number, event_id):
    """Remove the associated file event from the case, by event ID."""
    try:
        state.sdk.cases.file_events.delete(case_number, event_id)
    except Py42NotFoundError:
        raise Code42CLIError("Invalid case-number or event-id.")


@file_events.group(cls=OrderedGroup)
@sdk_options(hidden=True)
def bulk(state):
    """Tools for executing bulk case file-event actions."""
    pass


FILE_EVENTS_HEADERS = [
    "number",
    "event_id",
]

case_file_events_generate_template = generate_template_cmd_factory(
    group_name="file_events",
    commands_dict={"add": FILE_EVENTS_HEADERS, "remove": FILE_EVENTS_HEADERS},
)
bulk.add_command(case_file_events_generate_template)


@bulk.command(
    name="add",
    help="Bulk associate file events to cases using a CSV file with "
    f"format: {','.join(FILE_EVENTS_HEADERS)}.",
)
@read_csv_arg(headers=FILE_EVENTS_HEADERS)
@sdk_options()
def bulk_add(state, csv_rows):
    sdk = state.sdk

    def handle_row(number, event_id):
        sdk.cases.file_events.add(number, event_id)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Associating file events to cases:",
    )


@bulk.command(
    name="remove",
    help="Bulk remove the file event association from cases using a CSV file with "
    f"format: {','.join(FILE_EVENTS_HEADERS)}.",
)
@read_csv_arg(headers=FILE_EVENTS_HEADERS)
@sdk_options()
def bulk_remove(state, csv_rows):
    sdk = state.sdk

    def handle_row(number, event_id):
        sdk.cases.file_events.delete(number, event_id)

    run_bulk_process(
        handle_row,
        csv_rows,
        progress_label="Removing the file event association from cases:",
    )
